import os

from kyoka.utils import value_function_check
from kyoka.policy import GreedyPolicy
from kyoka.value_function import BaseTabularActionValueFunction, BaseApproxActionValueFunction
from kyoka.algorithm.rl_algorithm import BaseRLAlgorithm


class QLearning(BaseRLAlgorithm):
    """Basic "off-policy" Temporal-Difference Learning method.

    "off-policy" indicates that this method uses different policy when
    "select action during the episode" and "create backup target".
    (backup target is the target value used when training value function)

    QLearning uses GreedyPolicy to "create backup target".
    And you can choose policy to use to "select action during the episode".
    (Most of the case, soft-policy(ex. epsilon-greedy) is good choice.)

    Algorithm is implemented based on the book "Reinforcement Learning: An Introduction"
    (reference : https://webdocs.cs.ualberta.ca/~sutton/book/bookdraft2016sep.pdf)

    - Algorithm -
    Initialize
        T  <- your RL task
        PI <- policy used to "select action during the episode"
        Q  <- action value function
        a  <- learning rate (alpha)
        g  <- discounting factor (gamma)
    Repeat until computational budge runs out:
        S <- generate initial state of task T
        A <- choose action at S by following policy PI
        Repeat until S is terminal state:
            S' <- next state of S after taking action A
            R <- reward gained by taking action A at state S
            A' <- next action at S' by following policy PI
            GA <- greedy action at S' under action value function Q
            Q(S, A) <- Q(S, A) + a * [ R + g * Q(S', GA) - Q(S, A)]
            S, A <- S', A'
    """

    def __init__(self, alpha=0.1, gamma=0.9):
        """
        Args:
            alpha: learning rate. default=0.1. 0 < alpha <= 1
            gamma: discounting factor. default=0.9. 0 < gamma <= 1
        """
        self.alpha = alpha
        self.gamma = gamma

    def setup(self, task, policy, value_function):
        validate_value_function(value_function)
        super(QLearning, self).setup(task, policy, value_function)
        self.greedy_policy = GreedyPolicy()

    def run_gpi_for_an_episode(self, task, policy, value_function):
        state = task.generate_initial_state()
        action = policy.choose_action(task, value_function, state)
        while not task.is_terminal_state(state):
            next_state = task.transit_state(state, action)
            next_action = choose_action(task, policy, value_function, next_state)
            reward = task.calculate_reward(next_state)
            greedy_action = choose_action(task, self.greedy_policy, value_function, next_state)
            greedy_Q_value = predict_value(value_function, next_state, greedy_action)
            backup_target = reward + self.gamma * greedy_Q_value
            value_function.backup(state, action, backup_target, self.alpha)
            state, action = next_state, next_action

class QLearningTabularActionValueFunction(BaseTabularActionValueFunction):
    """Tabular action value function for QLearning.

    Backup target(TD error) passed from QLearning is "R + g * Q(S', GA)".
    So backup is done by following next equation.
        Q(S, A) <- Q(S, A) + a * [ R + g * Q(S', GA) - Q(S, A)]
    """

    def define_save_file_prefix(self):
        return "q_learning"

    def backup(self, state, action, backup_target, alpha):
        Q_value = self.predict_value(state, action)
        new_Q_value = Q_value + alpha * (backup_target - Q_value)
        self.insert_value_into_table(self.table, state, action, new_Q_value)

class QLearningApproxActionValueFunction(BaseApproxActionValueFunction):
    """Approximation action value function for QLearning.
    There is no additional method from base class to use QLearning.

    Backup target(TD error) passed from QLearning is "R + g * Q(S', GA)".
    So backup should be done to approximate next update.
        Q(S, A) <- Q(S, A) + a * [ R + g * Q(S', GA) - Q(S, A)]
    """
    pass

ACTION_ON_TERMINAL_FLG = "action_on_terminal"

def choose_action(task, policy, value_function, state):
    if task.is_terminal_state(state):
        return ACTION_ON_TERMINAL_FLG
    else:
        return policy.choose_action(task, value_function, state)

def predict_value(value_function, next_state, next_action):
    if ACTION_ON_TERMINAL_FLG == next_action:
        return 0
    else:
        return value_function.predict_value(next_state, next_action)

def validate_value_function(value_function):
    value_function_check("QLearning",
            [QLearningTabularActionValueFunction, QLearningApproxActionValueFunction],
            value_function)




import os

from kyoka.utils import build_not_implemented_msg, pickle_data, unpickle_data


class BaseActionValueFunction(object):
    """Base class of tabular and approximation action value function.

    The responsibility of action value function is to impelement two methods
    - predict_value
    - backup
    """

    def predict_value(self, state, action):
        """Predict the value of passed state-action pair.
        Returns:
            value: predicted value of passed state-action pair
        """
        err_msg = build_not_implemented_msg(self, "predict_value")
        raise NotImplementedError(err_msg)

    def backup(self, state, action, backup_target, alpha):
        """Update the value of passed state-action pair
        Args:
            state : state of state-action pair to update the value
            action : action of state-action pair to update the value
            backup_target : update the value by using this target which created from RL algorithm
            alpha : learning parameter passed from RL algorithm
        """
        err_msg = build_not_implemented_msg(self, "backup")
        raise NotImplementedError(err_msg)

    def setup(self):
      pass

    def save(self, save_dir_path):
      pass

    def load(self, load_dir_path):
      pass


class BaseTabularActionValueFunction(BaseActionValueFunction):
    """Base class of tabular value function used in RL algorithms

    property:
        table : the table to store the values. This property is initialized
                in setup method by using "generate_initial_table" method.
    """

    BASE_SAVE_FILE_NAME = "table_action_value_function_data.pickle"

    def generate_initial_table(self):
        """Initialize table to store the values of state-action pairs.
        Returns:
            table: this table is passed to "fetch_value_from_table" and
                   "insert_value_into_table" methods.
        """
        err_msg = build_not_implemented_msg(self, "generate_initial_table")
        raise NotImplementedError(err_msg)

    def fetch_value_from_table(self, table, state, action):
        """Define how to fetch the value of state-action pair from table.
        Args:
            table : current table object which initialzed by "generate_initial_table" method
            state : state of state-action pair to fetch the value from table
            action : action of state-action pair to fetch the value from table
        Returns:
            value : the value of state-action pair fetched from the table
        """
        err_msg = build_not_implemented_msg(self, "fetch_value_from_table")
        raise NotImplementedError(err_msg)

    def insert_value_into_table(self, table, state, action, new_value):
        """how to insert the new_item into table indexed by state-action pair

        This method directly update passed table by inserting new_value.
        (so thie method causes side-effect through table object)

        Args:
            table : table to insert the value (initialized by "generate_initial_table" method)
            state : state of state-action pair to index where to insert the new_value
            action: action of state-action pair to index where to insert the new_value
            new_value : new_value to insert into the table
        Returns:
            nothing : because directly update passed table object
        """
        err_msg = build_not_implemented_msg(self, "insert_value_into_table")
        raise NotImplementedError(err_msg)

    def define_save_file_prefix(self):
        """
        If you return "boo" then "self.save("some_dir")" will create
        "some_dir/boo_table_action_value_function_data.pickle"
        """
        return ""

    def setup(self):
        self.table = self.generate_initial_table()

    def predict_value(self, state, action):
        return self.fetch_value_from_table(self.table, state, action)

    def save(self, save_dir_path):
        pickle_data(self._gen_table_data_file_path(save_dir_path), self.table)

    def load(self, load_dir_path):
        file_path = self._gen_table_data_file_path(load_dir_path)
        if not os.path.exists(file_path):
            raise IOError('The saved data of "TableActionValueFunction"' +
                          ' is not found on [ %s ]' % load_dir_path)
        self.table = unpickle_data(file_path)

    def _gen_table_data_file_path(self, dir_path):
        return os.path.join(dir_path, self._gen_table_data_file_name())

    def _gen_table_data_file_name(self):
        prefix = self.define_save_file_prefix()
        if len(prefix) != 0: prefix += "_"
        return prefix + self.BASE_SAVE_FILE_NAME


class BaseApproxActionValueFunction(BaseActionValueFunction):
    """Base class of approximation value function

    Child class needs to implement following 3 methods for
    "predict_value" and "backup" which is necessary for action value function.

    - construct_features : transform state to feature representation
    - approx_predict_value : predict value by using feature representation
    - approx_backup : backup target valie by using feature representation
    """

    def predict_value(self, state, action):
        return self.approx_predict_value(self.construct_features(state, action))

    def backup(self, state, action, backup_target, alpha):
        self.approx_backup(self.construct_features(state, action), backup_target, alpha)

    def construct_features(self, state, action):
        """Transform state to compact feature representation
        Args:
            state: state of state-action pair to transform
            action: action of state-action pair to transform
        Returns:
            features: features which represents passed state-action pair
        """
        err_msg = build_not_implemented_msg(self, "construct_features")
        raise NotImplementedError(err_msg)

    def approx_predict_value(self, features):
        """Predict value by using feature representation of state-action pair
        Args:
            features: transformed by "construct_features" method
        Returns:
            value : predict the value of state-action pair by using features
        """
        err_msg = build_not_implemented_msg(self, "approx_predict_value")
        raise NotImplementedError(err_msg)

    def approx_backup(self, features, backup_target, alpha):
        """Update value by using feature representation of state-action pair
        Args:
            features: transformed by "construct_features" method
            backup_target : update the value by using this target which created from RL algorithm
            alpha : learning parameter passed from RL algorithm
        """
        err_msg = build_not_implemented_msg(self, "approx_backup")
        raise NotImplementedError(err_msg)




import os

from kyoka.utils import value_function_check
from kyoka.value_function import BaseTabularActionValueFunction, BaseApproxActionValueFunction
from kyoka.algorithm.rl_algorithm import BaseRLAlgorithm, generate_episode


class Sarsa(BaseRLAlgorithm):
    """Basic "on-policy" Temporal-Difference Learning method.

    "on-policy" indicates that this method uses same policy when
    "select action during the episode" and "create backup target".
    (backup target is the target value used when training value function)

    Algorithm is implemented based on the book "Reinforcement Learning: An Introduction"
    (reference : https://webdocs.cs.ualberta.ca/~sutton/book/bookdraft2016sep.pdf)

    - Algorithm -
    Initialize
        T  <- your RL task
        PI <- policy used in the algorithm
        Q  <- action value function
        a  <- learning rate (alpha)
        g  <- discounting factor (gamma)
    Repeat until computational budge runs out:
        S <- generate initial state of task T
        A <- choose action at S by following policy PI
        Repeat until S is terminal state:
            S' <- next state of S after taking action A
            R <- reward gained by taking action A at state S
            A' <- next action at S' by following policy PI
            Q(S, A) <- Q(S, A) + a * [ R + g * Q(S', A') - Q(S, A)]
            S, A <- S', A'
    """

    def __init__(self, alpha=0.1, gamma=0.9):
        """
        Args:
            alpha: learning rate. default=0.1. 0 < alpha <= 1
            gamma: discounting factor. default=0.9. 0 < gamma <= 1
        """
        self.alpha = alpha
        self.gamma = gamma

    def setup(self, task, policy, value_function):
        validate_value_function(value_function)
        super(Sarsa, self).setup(task, policy, value_function)

    def run_gpi_for_an_episode(self, task, policy, value_function):
        state = task.generate_initial_state()
        action = policy.choose_action(task, value_function, state)
        while not task.is_terminal_state(state):
            next_state = task.transit_state(state, action)
            next_action = choose_action(task, policy, value_function, next_state)
            reward = task.calculate_reward(next_state)
            next_Q_value = predict_value(value_function, next_state, next_action)
            backup_target = reward + self.gamma * next_Q_value
            value_function.backup(state, action, backup_target, self.alpha)
            state, action = next_state, next_action

class SarsaTabularActionValueFunction(BaseTabularActionValueFunction):
    """Tabular action value function for Sarsa.

    Backup target(TD error) passed from Sarsa is "R + g * Q(S', A')".
    So backup is done by following next equation.
        Q(S, A) <- Q(S, A) + a * [ R + g * Q(S', A') - Q(S, A)]
    """

    def define_save_file_prefix(self):
        return "sarsa"

    def backup(self, state, action, backup_target, alpha):
        Q_value = self.predict_value(state, action)
        new_Q_value = Q_value + alpha * (backup_target - Q_value)
        self.insert_value_into_table(self.table, state, action, new_Q_value)

class SarsaApproxActionValueFunction(BaseApproxActionValueFunction):
    """Approximation action value function for Sarsa.
    There is no additional method from base class to use QLearning.

    Backup target(TD error) passed from Sarsa is "R + g * Q(S', A')".
    So backup should be done to approximate next update.
        Q(S, A) <- Q(S, A) + a * [ R + g * Q(S', A') - Q(S, A)]
    """
    pass

ACTION_ON_TERMINAL_FLG = "action_on_terminal"

def choose_action(task, policy, value_function, state):
    if task.is_terminal_state(state):
        return ACTION_ON_TERMINAL_FLG
    else:
        return policy.choose_action(task, value_function, state)

def predict_value(value_function, next_state, next_action):
    if ACTION_ON_TERMINAL_FLG == next_action:
        return 0
    else:
        return value_function.predict_value(next_state, next_action)

def validate_value_function(value_function):
    value_function_check("Sarsa",
            [SarsaTabularActionValueFunction, SarsaApproxActionValueFunction],
            value_function)




from kyoka.task import BaseTask

class MazeTask(BaseTask):

    UP    = 0
    DOWN  = 1
    RIGHT = 2
    LEFT  = 3

    def read_maze(self, filepath):
        with open(filepath, 'rb') as f:
            self.maze = [line.rstrip() for line in f.readlines()]
            self.validate_maze(self.maze)

    def get_maze_shape(self):
        width, height = len(self.maze[0]), len(self.maze)
        return (height, width)

    def generate_initial_state(self):
        return self.__find_cell('S')

    def is_terminal_state(self, state):
        row, col = state
        return 'G' == self.maze[row][col]

    def transit_state(self, state, action):
        row, col = state
        height, width = self.get_maze_shape()
        if action == self.UP:
            row = max(0, row-1)
        elif action == self.DOWN:
            row = min(height-1, row+1)
        elif action == self.RIGHT:
            col= min(width-1, col+1)
        elif action == self.LEFT:
            col = max(0, col-1)
        if 'X' != self.maze[row][col]:
            return (row, col)
        else:
            return state

    def generate_possible_actions(self, state):
        return [self.UP, self.DOWN, self.RIGHT, self.LEFT]

    def calculate_reward(self, state):
        row, col = state
        return 1 if 'G' == self.maze[row][col] else 0


    def validate_maze(self, maze):
        cells = reduce(lambda x, y: x+y, maze)
        start_count = cells.count('S')
        goal_count = cells.count('G')
        if start_count != 1 or goal_count != 1:
            raise ValueError("Invalid maze passed. reason: start_count=%d, goal_count=%d")

    def generate_terminal_state(self):
        return self.__find_cell('G')

    def __find_cell(self, target):
        for row in range(len(self.maze)):
            for col in range(len(self.maze[0])):
                if self.maze[row][col] == target: return (row, col)




import os
import time

from utils import build_not_implemented_msg


class BaseCallback(object):
    """Base class for creating new GPI callback as you like.

    The following callback methods are called from BaseRLAlgorithm.run_gpi_for_an_episode
    in proper timing.

    before_gpi_start : called before start training
    before_update : called before each call of RLalgorithm.run_gpi_for_an_episode
    after_update : called after each call of RLalgorithm.run_gpi_for_an_episode
    after_gpi_finish : called when training is finished
    interrupt_gpi : called after each update is finished to judge if finish the training
    """

    def before_gpi_start(self, task, value_function):
        pass

    def before_update(self, iteration_count, task, value_function):
        pass

    def after_update(self, iteration_count, task, value_function):
        pass

    def after_gpi_finish(self, task, value_function):
        pass

    def interrupt_gpi(self, iteration_count, task, value_function):
        """Return True if you want to stop the training."""
        return False

    def define_log_tag(self):
        """Define tag string which displayed with log message.

        Ex. if you return "MyTag" here then
            self.log("some message") would output
            "[MyTag] some message" on your console
        """
        return self.__class__.__name__

    @property
    def tag(self):
        return self.define_log_tag()

    def log(self, message):
        if message and len(message) != 0:
            print "[%s] %s" % (self.tag, message)

class BasePerformanceWatcher(BaseCallback):
    """Utility class to execute some calculation with intermediate result of training.
    """

    def setUp(self, task, value_function):
        pass

    def tearDown(self, task, value_function):
        pass

    def define_performance_test_interval(self):
        """Define interval to execute "run_performance_test".

        For example, if you return 1 then "run_performance_test" is called after
        every update of training.
        """
        err_msg = build_not_implemented_msg(self, "define_performance_test_interval")
        raise NotImplementedError(err_msg)

    def run_performance_test(self, task, value_function):
        """Define some calculation to see how good current value_function works.

        Args:
            task: Task instance which is used in training
            value_function: Value function which is in the middle of training.
        """
        err_msg = build_not_implemented_msg(self, "run_performance_test")
        raise NotImplementedError(err_msg)

    def define_log_message(self, iteration_count, task, value_function, test_result):
        base_msg = "Performance test result : %s (nb_iteration=%d)"
        return base_msg % (test_result, iteration_count)


    def before_gpi_start(self, task, value_function):
        self.performance_log = []
        self.test_interval = self.define_performance_test_interval()
        self.setUp(task, value_function)

    def after_update(self, iteration_count, task, value_function):
        if iteration_count % self.test_interval == 0:
            result = self.run_performance_test(task, value_function)
            self.performance_log.append(result)
            message = self.define_log_message(iteration_count, task, value_function, result)
            self.log(message)

    def after_gpi_finish(self, task, value_function):
        self.tearDown(task, value_function)

class EpsilonAnnealer(BaseCallback):
    """Callback to decay epsiron of EpsironGreedyPolicy during training."""

    def __init__(self, epsilon_greedy_policy):
        """
        Args:
            epsilon_greedy_policy: target to execute epsilon annealing.
        """
        self.policy = epsilon_greedy_policy
        self.anneal_finished = False

    def define_log_tag(self):
        return "EpsilonGreedyAnnealing"

    def before_gpi_start(self, _task, _value_function):
        start_msg = "Anneal epsilon from %s to %s." % (self.policy.eps, self.policy.min_eps)
        self.log(start_msg)

    def after_update(self, iteration_count, _task, _value_function):
        self.policy.anneal_eps()
        if not self.anneal_finished and self.policy.eps == self.policy.min_eps:
            self.anneal_finished = True
            finish_msg = "Annealing has finished at %d iteration." % iteration_count
            self.log(finish_msg)

class LearningRecorder(BaseCallback):
    """Callback to save intermediate result of training.

    If you set
        root_save_dir_path=training_results
        save_interval=100
    After 250 iteration of training, "algorithm.save(training_results)" is called twice.
    So training_results directory would have two items like below

    training_results/after_100_iteration/...
                    /after_200_iteration/...
    """

    def __init__(self, algorithm, root_save_dir_path, save_interval):
        """
        Args:
            algorithm: the RL algorithm which will be used in training.
            root_save_dir_path: save method is executed under this path of directory.
            save_interval: interval of training to execute save method.
        """
        self.algorithm = algorithm
        self.root_save_dir_path = root_save_dir_path
        self.save_interval = save_interval

    def before_gpi_start(self, _task, _value_function):
        if not os.path.exists(self.root_save_dir_path):
            err_msg = "Directory [ %s ] not found which you passed to LearningRecorder."
            raise Exception(err_msg  % self.root_save_dir_path)
        base_msg = 'Your algorithm will be saved after each %d iteration on directory [ %s ].'
        self.log(base_msg % (self.save_interval, self.root_save_dir_path))

    def after_update(self, iteration_count, _task, _value_function):
        if iteration_count % self.save_interval == 0:
            dir_name = self.define_checkpoint_save_dir_name(iteration_count)
            save_path = os.path.join(self.root_save_dir_path, dir_name)
            os.mkdir(save_path)
            self.algorithm.save(save_path)
            base_msg = "Saved algorithm after %d iteration at [ %s ]."
            self.log(base_msg % (iteration_count, save_path))

    def after_gpi_finish(self, task, value_function):
        dir_name = self.define_finish_save_dir_name()
        save_path = os.path.join(self.root_save_dir_path, dir_name)
        os.mkdir(save_path)
        self.algorithm.save(save_path)

    def define_checkpoint_save_dir_name(self, iteration_count):
        return "after_%d_iteration" % iteration_count

    def define_finish_save_dir_name(self):
        return "gpi_finished"

class BaseFinishRule(BaseCallback):
    """Base class to define the rule to stop the training.

    Child class needs to implement following 3 methods.

    check_condition : return True if you want  to stop the training.
    generate_start_message : the string retured here is logged when start training.
    generate_finish_message : the string returned here is logged after training finished.
    """

    def check_condition(self, iteration_count, task, value_function):
        """Return True if you want to stop the training"""
        err_msg = build_not_implemented_msg(self, "check_condition")
        raise NotImplementedError(err_msg)

    def generate_start_message(self):
        err_msg = build_not_implemented_msg(self, "generate_start_message")
        raise NotImplementedError(err_msg)

    def generate_finish_message(self, iteration_count):
        err_msg = build_not_implemented_msg(self, "generate_finish_message")
        raise NotImplementedError(err_msg)

    def before_gpi_start(self, task, value_function):
        self.log(self.generate_start_message())

    def interrupt_gpi(self, iteration_count, task, value_function):
        finish_iteration = self.check_condition(iteration_count, task, value_function)
        if finish_iteration: self.log(self.generate_finish_message(iteration_count))
        return finish_iteration

class ManualInterruption(BaseFinishRule):
    """Callback to stop the training by file-base communication

    How to stop the training manually
    1. Share the path of a file with this callback
    2. Start the training with this callback.
    3. Write target word (default is "stop") on the shared file.
    4. Callback will find the target word from the shared file and stop the training.
    """

    TARGET_WARD = "stop"

    def __init__(self, monitor_file_path, watch_interval=30):
        """
        Args:
            monitor_file_path: path of file to send stop command.
            watch_interval: the interval to check the file in monitor_file_path
        """
        self.monitor_file_path = monitor_file_path
        self.watch_interval = watch_interval

    def check_condition(self, _iteration_count, _task, _value_function):
        current_time = time.time()
        if current_time - self.last_check_time >= self.watch_interval:
            self.last_check_time = current_time
            return self.__order_found_in_monitoring_file(self.monitor_file_path, self.TARGET_WARD)
        else:
            return False

    def generate_start_message(self):
        self.last_check_time = time.time()
        base_first_msg ='Write word "%s" on file "%s" will finish the GPI'
        base_second_msg = "(Stopping GPI may take about %s seconds. Because we check target file every %s seconds.)"
        first_msg = base_first_msg % (self.TARGET_WARD, self.monitor_file_path)
        second_msg = base_second_msg % (self.watch_interval, self.watch_interval)
        return "\n".join([first_msg, second_msg])

    def generate_finish_message(self, iteration_count):
        base_msg = "Interrupt GPI after %d iterations because interupption order found in [ %s ]."
        return base_msg % (iteration_count, self.monitor_file_path)

    def __order_found_in_monitoring_file(self, filepath, target_word):
        return os.path.isfile(filepath) and self.__found_target_ward_in_file(filepath, target_word)

    def __found_target_ward_in_file(self, filepath, target_word):
        search_word = lambda src, target: target in src
        src = self.__read_data(filepath)
        return search_word(src, target_word) if src else False

    def __read_data(self, filepath):
        with open(filepath, 'rb') as f: return f.read()

class WatchIterationCount(BaseFinishRule):
    """Finish the training after specified number of training."""

    def __init__(self, target_count, verbose=1):
        """
        Args
            target_count : stop the training after this number of iteration.
            verbose : if verbose > 0 then log is activated.
        """
        self.target_count = target_count
        self.start_time = self.last_update_time = 0
        self.verbose = verbose

    def define_log_tag(self):
        return "Progress"

    def check_condition(self, iteration_count, task, value_function):
        return iteration_count >= self.target_count

    def generate_start_message(self):
        self.start_time = self.last_update_time = time.time()
        return "Start GPI iteration for %d times" % self.target_count

    def generate_finish_message(self, iteration_count):
        base_msg = "Completed GPI iteration for %d times. (total time: %ds)"
        return base_msg % (iteration_count, time.time() - self.start_time)

    def before_update(self, iteration_count, task, value_function):
        super(WatchIterationCount, self).before_update(iteration_count, task, value_function)
        self.last_update_time = time.time()

    def after_update(self, iteration_count, task, value_function):
        super(WatchIterationCount, self).after_update(iteration_count, task, value_function)
        if self.verbose > 0:
            current_time = time.time()
            msg = "Finished %d / %d iterations (%.1fs)" % (
                    iteration_count, self.target_count,
                    current_time - self.last_update_time)
            self.last_update_time = current_time
            self.log(msg)




import os

from nose.tools import raises

from examples.maze.task import MazeTask
from tests.base_unittest import BaseUnitTest


class MazeTaskTest(BaseUnitTest):

  def setUp(self):
    self.task = MazeTask()
    self.task.read_maze(get_sample_maze_path())

  def test_read_maze(self):
    self.eq('S', self.task.maze[0][1])
    self.eq('G', self.task.maze[1][3])
    self.eq('-', self.task.maze[0][0])
    self.eq('X', self.task.maze[1][2])

  def test_get_maze_shape(self):
    self.eq((3,4), self.task.get_maze_shape())

  def test_generate_inital_state(self):
    state = self.task.generate_initial_state()
    self.eq((0,1), state)

  @raises(ValueError)
  def test_raise_error_when_no_start_maze_passed(self):
    no_start_maze = [["-", "G"]]
    self.task.validate_maze(no_start_maze)

  @raises(ValueError)
  def test_raise_error_when_no_goal_maze_passed(self):
    no_goal_maze = [["-", "S"]]
    self.task.validate_maze(no_goal_maze)

  def test_terminal_state_check_on_start(self):
    start = (0,1)
    self.false(self.task.is_terminal_state(start))

  def test_terminal_state_check_on_goal(self):
    goal = (1, 3)
    self.true(self.task.is_terminal_state(goal))

  def test_transit_state_from_initial_state(self):
    transition_info = [
        (self.task.UP, (0, 1)),
        (self.task.LEFT, (0, 0)),
        (self.task.LEFT, (0, 0)),
        (self.task.DOWN, (1, 0)),
        (self.task.RIGHT, (1, 1)),
        (self.task.RIGHT, (1, 1)),
        (self.task.DOWN, (2, 1)),
        (self.task.DOWN, (2, 1)),
        (self.task.RIGHT, (2, 2)),
        (self.task.RIGHT, (2, 3)),
        (self.task.RIGHT, (2, 3)),
        (self.task.UP, (1, 3))
    ]
    state = (0,1)
    for action, expected in transition_info:
      state = self.__transit_and_check(state, action, expected)
    self.true(self.task.is_terminal_state(state))

  def test_generate_possible_actions(self):
    state = "dummy"
    expected = [self.task.UP, self.task.DOWN, self.task.RIGHT, self.task.LEFT]
    self.eq(expected, self.task.generate_possible_actions(state))

  def test_calculate_reward(self):
    start, goal, empty, block = (0, 1), (1, 3), (0, 0), (1, 2)
    self.eq(0, self.task.calculate_reward(start))
    self.eq(1, self.task.calculate_reward(goal))
    self.eq(0, self.task.calculate_reward(empty))
    self.eq(0, self.task.calculate_reward(block))


  def __transit_and_check(self, state, action, expected):
    next_state = self.task.transit_state(state, action)
    self.eq(expected, next_state)
    return next_state

def get_sample_maze_path():
    return os.path.join(os.path.dirname(__file__), "sample_maze.txt")




from kyoka.utils import build_not_implemented_msg
from kyoka.policy import EpsilonGreedyPolicy
from kyoka.callback import EpsilonAnnealer, WatchIterationCount


def generate_episode(task, policy, value_function):
    """Utility method to generate an episode by following passed
    policy and value_function

    Args:
        task: Task object which represents some RL problem
        policy : generate episode by following this policy
        value_function : choose action by following this value function
    Returns:
        episode : Array of experience (tuple of state, action, next_state, reward).
                  Last item(experience) of next_state(3rd element of experience)
                  must be terminal state.
    """
    state = task.generate_initial_state()
    episode = []
    while not task.is_terminal_state(state):
        action = policy.choose_action(task, value_function, state)
        next_state = task.transit_state(state, action)
        reward = task.calculate_reward(next_state)
        episode.append((state, action, next_state, reward))
        state = next_state
    return episode

class BaseRLAlgorithm(object):
    """Base class for all of offline RL algorithms.

    "run_gpi" is the method for training (learning value function).
    But its core logic of how to update value function is delegated to
    abstract method "run_gpi_for_an_episode".
    So the role of child class is implement this abstract method.

    Do not forget to call setup method before calling "run_gpi" method.
    (If you forget, Exception is raised on "run_gpi" call.)
    """

    def setup(self, task, policy, value_function):
        """Set items which necesarry for training"""
        self.task = task
        self.value_function = value_function
        self.value_function.setup()
        self.policy = policy

    def save(self, save_dir_path):
        """Save value function and state of algorithm if exist.
        Args:
            save_dir_path: items are saved under this directory
        """
        self.value_function.save(save_dir_path)
        self.save_algorithm_state(save_dir_path)

    def load(self, load_dir_path):
        """load saved items from directory where save method executed.
        Args:
            load_dir_path: this should be the path you passed save method
                           as save_dir_path
        """
        self.value_function.load(load_dir_path)
        self.load_algorithm_state(load_dir_path)

    def save_algorithm_state(self, save_dir_path):
        """If algorithm uses some state variables in the training save it.
        Args:
            save_dir_path: items are saved under this directory
        """
        pass

    def load_algorithm_state(self, load_dir_path):
        """load saved state of algorithm from directory where
        "save_algorithm_state" method executed.
        Args:
            load_dir_path: this should be the path you passed
                           "save_algorithm_state" as save_dir_path
        """
        pass

    def run_gpi_for_an_episode(self, task, policy, value_function):
        """Define how to update value function for an episode."""
        err_msg = build_not_implemented_msg(self, "run_gpi_for_an_episode")
        raise NotImplementedError(err_msg)

    def run_gpi(self, nb_iteration, callbacks=None, verbose=1):
        """Run GPI(Generalized Policy Iteration) to improve value function
        Args:
            nb_iteration: run GPI for nb_iteration times
            callbacks: These callbacks are invoked during GPI.
                       callback objects (objects which inherit
                       "kyoka.callbacks.BaseCallback").
                       Array of callback objects or a callback object is acceptable.
            verbose: verbose > 0 logs progress of training
        """
        self.__check_setup_call()
        default_finish_rule = WatchIterationCount(nb_iteration, verbose)
        callbacks = self.__setup_callbacks(default_finish_rule, callbacks)
        [callback.before_gpi_start(self.task, self.value_function) for callback in callbacks]

        iteration_counter = 1
        while True:
            [callback.before_update(iteration_counter, self.task, self.value_function) for callback in callbacks]
            self.run_gpi_for_an_episode(self.task, self.policy, self.value_function)
            [callback.after_update(iteration_counter, self.task, self.value_function) for callback in callbacks]
            for finish_rule in callbacks:
                if finish_rule.interrupt_gpi(iteration_counter, self.task, self.value_function):
                    [callback.after_gpi_finish(self.task, self.value_function) for callback in callbacks]
                    if finish_rule != default_finish_rule:
                        default_finish_rule.log(default_finish_rule.generate_finish_message(iteration_counter))
                    return
            iteration_counter += 1


    def __check_setup_call(self):
        """Raise exception with message if run_gpi is called without setup"""
        if not all([hasattr(self, attr) for attr in ["task", "value_function", "policy"]]):
            raise Exception('You need to call "setup" method before calling "run_gpi" method.')

    def __setup_callbacks(self, default_finish_rule, user_callbacks):
        """Add EpsilonAnnealer if needed and put default_callback
        at top of callback list.
        """
        user_callbacks = self.__wrap_item_if_single(user_callbacks)
        default_callbacks = [default_finish_rule]
        if isinstance(self.policy, EpsilonGreedyPolicy) and self.policy.do_annealing:
            default_callbacks.append(EpsilonAnnealer(self.policy))
        return default_callbacks + user_callbacks

    def __wrap_item_if_single(self, item):
        """Make single callback object as single callback list"""
        if item is None: item = []
        if not isinstance(item, list): item = [item]
        return item




from utils import build_not_implemented_msg


class BaseTask(object):
    """Base class to represent your problem as RL task

    RL algorithm requires following 5 methods to learn ActionValueFunction from some task.
    - generate_initial_state
    - is_terminal_state
    - transit_state
    - generate_possible_actions
    - calculate_reward
    """

    def generate_inital_state(self):
        """Define start state of your problem.
        Returns:
            initial_state: representation of initial state of your problem.
        """
        err_msg = build_not_implemented_msg(self, "generate_inital_state")
        raise NotImplementedError(err_msg)

    def is_terminal_state(self, state):
        """Define when is the finish of episode (your problem)
        Returns:
            is_terminal(boolean): True if your problem is finished in the state.
        """
        err_msg = build_not_implemented_msg(self, "is_terminal_state")
        raise NotImplementedError(err_msg)

    def transit_state(self, state, action):
        """Define the rule of state transition in your problem
        Args:
            state: current state which agent exists.
            action: action choosed in current state
        Returns:
            next_state: next state after applied passed action to current state
        """
        err_msg = build_not_implemented_msg(self, "transit_state")
        raise NotImplementedError(err_msg)

    def generate_possible_actions(self, state):
        """Define what action is possible in each state
        Args:
            state: we want to know legal actions on this state
        Returns:
            actions: actions which is legal in passed state
        """
        err_msg = build_not_implemented_msg(self, "generate_possible_actions")
        raise NotImplementedError(err_msg)

    def calculate_reward(self, state):
        """Define reward of each state
        Args:
            state: we want to know the reward of this state
        Returns:
            reward(scalar): the reward of passed state
        """
        err_msg = build_not_implemented_msg(self, "calculate_reward")
        raise NotImplementedError(err_msg)




import os

from mock import Mock
from nose.tools import raises

from kyoka.algorithm.montecarlo import MonteCarlo, MonteCarloTabularActionValueFunction,\
        MonteCarloApproxActionValueFunction, validate_value_function
from kyoka.value_function import BaseActionValueFunction
from kyoka.policy import GreedyPolicy
from tests.base_unittest import BaseUnitTest
from tests.utils import generate_tmp_dir_path, setup_tmp_dir, teardown_tmp_dir


class MonteCarloTest(BaseUnitTest):

    def setUp(self):
        self.algo = MonteCarlo()

    def tearDown(self):
        cleanup_trash()

    def test_value_function_validation(self):
        validate_value_function(MonteCarloTabularActionValueFunction())
        validate_value_function(MonteCarloApproxActionValueFunction())
        with self.assertRaises(TypeError):
            self.algo.setup(BaseActionValueFunction())


    def test_run_gpi_for_an_episode(self):
        task = setup_stub_task()
        value_func = MonteCarloTabularActionValueFunctionImpl()
        policy = GreedyPolicy()
        self.algo.setup(task, policy, value_func)
        self.algo.run_gpi_for_an_episode(task, policy, value_func)
        expected = [(0, 1, 59, 1), (1, 2, 58, 1), (3, 4, 49, 1), (0, 0, 0, 0)]
        update_counter = value_func.update_counter
        for state, action, value, update_count in expected:
            self.eq(value, value_func.fetch_value_from_table(value_func.table, state, action))
            self.eq(update_count, value_func.fetch_value_from_table(update_counter, state, action))

    def test_integration(self):
        task = setup_stub_task()
        value_func = MonteCarloTabularActionValueFunctionImpl()
        policy = GreedyPolicy()
        self.algo.setup(task, policy, value_func)
        self.algo.run_gpi_for_an_episode(task, policy, value_func)
        setup_tmp_dir(__file__)
        self.algo.save(generate_tmp_dir_path(__file__))

        task.calculate_reward.side_effect = lambda state: state
        new_algo = MonteCarlo()
        new_algo.setup(task, policy, value_func)
        new_algo.load(generate_tmp_dir_path(__file__))
        new_algo.run_gpi_for_an_episode(task, policy, value_func)

        expected = [(0, 1, 35, 2), (1, 2, 34, 2), (3, 4, 28, 2), (0, 0, 0, 0)]
        update_counter = value_func.update_counter
        for state, action, value, update_count in expected:
            self.eq(value, value_func.fetch_value_from_table(value_func.table, state, action))
            self.eq(update_count, value_func.fetch_value_from_table(update_counter, state, action))

    def test_reward_discounting(self):
        no_discount = MonteCarlo()
        episode = [("s", "a", "ns", 4), ("s", "a", "ns", 2), ("s", "a", "ns", 1), ("s", "a", "ns", 8)]
        self.eq(15, no_discount._calculate_following_state_reward(0, episode))
        self.eq(11, no_discount._calculate_following_state_reward(1, episode))
        self.eq(9, no_discount._calculate_following_state_reward(2, episode))
        self.eq(8, no_discount._calculate_following_state_reward(3, episode))
        discount = MonteCarlo(gamma=0.9)
        self.eq(12.442, discount._calculate_following_state_reward(0, episode))
        self.eq(9.38, discount._calculate_following_state_reward(1, episode))
        self.eq(8.2, discount._calculate_following_state_reward(2, episode))
        self.eq(8, discount._calculate_following_state_reward(3, episode))



class MonteCarloTabularActionValueFunctionTest(BaseUnitTest):

    def setUp(self):
        self.func = MonteCarloTabularActionValueFunctionImpl()

    def tearDown(self):
        cleanup_trash()

    def test_setup(self):
        self.func.setup()
        self.eq(4, len(self.func.update_counter))

    def test_save_and_load(self):
        self.func.setup()
        self.func.update_counter[0][0] = 1
        setup_tmp_dir(__file__)
        self.func.save(generate_tmp_dir_path(__file__))
        new_func = MonteCarloTabularActionValueFunctionImpl()
        new_func.load(generate_tmp_dir_path(__file__))
        self.eq(1, new_func.update_counter[0][0])

    def test_backup(self):
        self.func.setup()
        self.func.backup(state=0, action=1, backup_target=2, alpha="dummy")
        self.func.backup(state=1, action=0, backup_target=3, alpha="dummy")
        self.func.backup(state=0, action=1, backup_target=4, alpha="dummy")
        self.eq(3, self.func.predict_value(state=0, action=1))
        self.eq(3, self.func.predict_value(state=1, action=0))
        self.eq(2, self.func.update_counter[0][1])
        self.eq(1, self.func.update_counter[1][0])

class MonteCarloApproxActionValueFunctionTest(BaseUnitTest):

    def setUp(self):
        self.func = MonteCarloApproxActionValueFunction()

    @raises(NotImplementedError)
    def test_error_msg_when_not_implement_construct_features(self):
        self.func.construct_features("dummy", "dummy")

    @raises(NotImplementedError)
    def test_error_msg_when_not_implement_approx_predict_value(self):
        self.func.approx_predict_value("dummy")

    @raises(NotImplementedError)
    def test_error_msg_when_not_implement_approx_backup(self):
        self.func.approx_backup("dummy", "dummy", "dummy")


class MonteCarloTabularActionValueFunctionImpl(MonteCarloTabularActionValueFunction):

    def generate_initial_table(self):
        return [[0 for i in range(50)] for i in range(4)]

    def fetch_value_from_table(self, table, state, action):
        return table[state][action]

    def insert_value_into_table(self, table, state, action, new_value):
        table[state][action] = new_value

def setup_stub_task():
    mock_task = Mock()
    mock_task.generate_initial_state.return_value = 0
    mock_task.is_terminal_state.side_effect = lambda state: state == 7
    mock_task.transit_state.side_effect = lambda state, action: state + action
    mock_task.generate_possible_actions.side_effect = lambda state: [state + 1]
    mock_task.calculate_reward.side_effect = lambda state: state**2
    return mock_task

def cleanup_trash():
    filenames = ["montecarlo_update_counter.pickle", "montecarlo_table_action_value_function_data.pickle"]
    teardown_tmp_dir(__file__, filenames)




import os

from nose.tools import raises
from mock import patch, Mock

from kyoka.utils import pickle_data, unpickle_data
from kyoka.policy import GreedyPolicy
from kyoka.algorithm.deep_q_learning import DeepQLearning,\
        DeepQLearningApproxActionValueFunction, ExperienceReplay
from tests.base_unittest import BaseUnitTest
from tests.utils import generate_tmp_dir_path, setup_tmp_dir, teardown_tmp_dir, NegativePolicy


class DeepQLearningTest(BaseUnitTest):

    def setUp(self):
        self.algo = DeepQLearning(gamma=0.1, N=3, C=3, minibatch_size=2, replay_start_size=2)
        self.algo.replay_memory.store_transition(2.5, 3, 25, 5)
        self.algo.replay_memory.store_transition(5.0, 7, 144, 4)
        self.task = setup_stub_task()
        self.policy = NegativePolicy()
        self.value_func = self.DeepQLearningApproxActionValueFunctionImpl()
        self.algo.setup(self.task, self.policy, self.value_func)

    def tearDown(self):
        cleanup_trash()

    def test_initialize_replay_memory(self):
        algo = DeepQLearning(gamma=0.1, N=3, C=3, minibatch_size=2, replay_start_size=2)
        task = setup_stub_task()
        policy = NegativePolicy()
        value_func = self.DeepQLearningApproxActionValueFunctionImpl(strict_mode=False)
        # Overrider terminal judge logic to avoid infinite episode by random policy
        task.is_terminal_state.side_effect = lambda state: state == 4 or state >= 100
        self.eq(0, len(algo.replay_memory.queue))
        algo.setup(task, policy, value_func)
        self.eq(2, len(algo.replay_memory.queue))

    def test_check_backup_minibatch_delivery_in_gpi(self):
        with patch('random.sample', side_effect=lambda lst, n: lst[len(lst)-n:]):
            self.algo.run_gpi_for_an_episode(self.task, self.policy, self.value_func)

        backup_minibatch_expected = [
                [(5.0, 7, 144), (0, 1, 1.6)],
                [(0, 1, 1.6), (1, 3, 16)]
        ]
        actual = [arg[0][0] for arg in self.value_func.q_network.train_on_minibatch.call_args_list]
        self.eq(backup_minibatch_expected, actual)
        self.value_func.q_hat_network.train_on_minibatch.assert_not_called()

    def test_update_value_function_reset_target_network(self):
        with patch('random.sample', side_effect=lambda lst, n: lst[-n:]):
            self.algo.run_gpi_for_an_episode(self.task, self.policy, self.value_func)
        self.eq(2, self.algo.reset_step_counter)
        self.eq("Q_hat_network_0", self.value_func.q_hat_network.name)
        self.algo.run_gpi_for_an_episode(self.task, self.policy, self.value_func)
        self.eq(0, self.algo.reset_step_counter)
        self.eq("Q_hat_network_1", self.value_func.q_hat_network.name)

    def test_save_and_load_algorithm_state(self):
        dir_path = generate_tmp_dir_path(__file__)
        file_path = os.path.join(dir_path, "dq_replay_memory.pickle")
        setup_tmp_dir(__file__)
        self.algo.save_algorithm_state(dir_path)
        self.true(os.path.exists(file_path))

        new_algo = DeepQLearning(replay_start_size=100)
        task = setup_stub_task()
        # Overrider terminal judge logic to avoid infinite episode by random policy
        task.is_terminal_state.side_effect = lambda state: state == 4 or state >= 100
        policy = NegativePolicy()
        value_func = self.DeepQLearningApproxActionValueFunctionImpl(strict_mode=False)
        new_algo.setup(task, policy, value_func)
        new_algo.load_algorithm_state(dir_path)

        # Validate algorithm's state
        self.eq(self.algo.gamma, new_algo.gamma)
        self.eq(self.algo.C, new_algo.C)
        self.eq(self.algo.minibatch_size, new_algo.minibatch_size)
        self.eq(self.algo.replay_start_size, new_algo.replay_start_size)
        self.eq(self.algo.reset_step_counter, new_algo.reset_step_counter)
        self.eq(task, new_algo.task)
        self.eq(policy, new_algo.policy)
        self.eq(value_func, new_algo.value_function)
        self.eq(self.algo.replay_memory.max_size, new_algo.replay_memory.max_size)
        self.eq(self.algo.replay_memory.queue, new_algo.replay_memory.queue)
        self.true(isinstance(new_algo.greedy_policy, GreedyPolicy))

       # Validate that loaded algorithm works like original one
        with patch('random.sample', side_effect=lambda lst, n: lst[len(lst)-n:]):
            new_algo.run_gpi_for_an_episode(self.task, self.policy, self.value_func)
        replay_memory_expected = [
                (5.0, 7, 144, 4),
                (0, 1, 1, 1),
                (1, 3, 16, 4)
        ]
        self.eq(replay_memory_expected, new_algo.replay_memory.queue)

    class DeepQLearningApproxActionValueFunctionImpl(DeepQLearningApproxActionValueFunction):

        def __init__(self, strict_mode=True):
            self.deepcopy_counter = 0
            self.strict_mode = strict_mode

        def initialize_network(self):
            mock_q_network = Mock(name="Q_network")
            mock_q_network.predict.side_effect = self.q_predict_scenario
            return mock_q_network

        def deepcopy_network(self, q_network):
            mock_q_hat_network = Mock(name="Q_hat_network")
            mock_q_hat_network.name = "Q_hat_network_%d" % self.deepcopy_counter
            mock_q_hat_network.predict.side_effect = self.q_hat_predict_scenario
            self.deepcopy_counter += 1
            return mock_q_hat_network

        def predict_value_by_network(self, network, state, action):
            return network.predict(state, action)

        def backup_on_minibatch(self, q_network, backup_minibatch):
            q_network.train_on_minibatch(backup_minibatch)

        def q_predict_scenario(self, state, action):
            if state == 0 and action == 1:
                return 1
            elif state == 0 and action == 2:
                return 2
            elif state == 1 and action == 2:
                return 4
            elif state == 1 and action == 3:
                return 3
            else:
                if self.strict_mode:
                    raise AssertionError("q_network received unexpected state-action pair (state=%s, action=%s)" % (state, action))
                else:
                    return 1

        def q_hat_predict_scenario(self, state, action):
            if state == 1 and action == 2:
                return 5
            elif state == 1 and action == 3:
                return 6
            else:
                if self.strict_mode:
                    raise AssertionError("q_hat_network received unexpected state-action pair (state=%s, action=%s)" % (state, action))
                else:
                    return 1

class DeepQLearningApproxValueFunctionTest(BaseUnitTest):

    def setUp(self):
        self.empty_func = DeepQLearningApproxActionValueFunction()
        self.func = self.DeepQLearningApproxActionValueFunctionImpl()
        self.func.setup()

    def tearDown(self):
        cleanup_trash()

    @raises(NotImplementedError)
    def test_initialize_network(self):
        self.empty_func.initialize_network()

    @raises(NotImplementedError)
    def test_deepcopy_network(self):
        self.empty_func.deepcopy_network("dummy")

    @raises(NotImplementedError)
    def test_predict_value_by_network(self):
        self.empty_func.predict_value_by_network("dummy", "dummy", "dummy")

    @raises(NotImplementedError)
    def test_backup_on_minibatch(self):
        self.empty_func.backup_on_minibatch("dummy", "dummy")

    @raises(NotImplementedError)
    def test_save_networks(self):
        self.empty_func.save_networks("dummy", "dummy", "dummy")

    @raises(NotImplementedError)
    def test_load_networks(self):
        self.empty_func.load_networks("dummy")

    def test_setup(self):
        self.eq(0, self.func.q_network)
        self.eq(1, self.func.q_hat_network)
        self.false(self.func.use_target_network_flg)

    def test_use_target_network(self):
        self.func.use_target_network(True)
        self.true(self.func.use_target_network_flg)
        self.func.use_target_network(False)
        self.false(self.func.use_target_network_flg)

    def test_calculate_value(self):
        self.eq(0, self.func.predict_value(2, 3))
        self.func.use_target_network(True)
        self.eq(6, self.func.predict_value(2, 3))

    def test_reset_target_network(self):
        self.func.q_network = 2
        self.func.reset_target_network()
        self.eq(3, self.func.q_hat_network)

    def test_save_and_load_networks(self):
        self.func.q_network = 2
        dir_path = generate_tmp_dir_path(__file__)
        setup_tmp_dir(__file__)
        self.func.save(dir_path)
        new_func = self.DeepQLearningApproxActionValueFunctionImpl()
        new_func.setup()
        new_func.load(dir_path)
        self.eq(self.func.q_network, new_func.q_network)
        self.eq(self.func.q_hat_network, new_func.q_hat_network)

    class DeepQLearningApproxActionValueFunctionImpl(DeepQLearningApproxActionValueFunction):

        def initialize_network(self):
            return 0

        def deepcopy_network(self, q_network):
            return q_network + 1

        def predict_value_by_network(self, network, state, action):
            return network * state * action

        def save_networks(self, q_network, q_hat_network, save_dir_path):
            pickle_data(gen_tmp_file_path(save_dir_path), (q_network, q_hat_network))

        def load_networks(self, load_dir_path):
            q_network, q_hat_network = unpickle_data(gen_tmp_file_path(load_dir_path))
            return q_network, q_hat_network


class ExperienceReplayTest(BaseUnitTest):

    def test_queue_size_upper_bound(self):
        er = ExperienceReplay(max_size=2)
        experiences = [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9 ,0, 1)]
        for e in experiences:
            er.store_transition(state=e[0], action=e[1], reward=e[2], next_state=e[3])

        self.neq(er.queue[0], experiences[0])
        self.eq(er.queue[0], experiences[1])
        self.eq(er.queue[1], experiences[2])

    def test_sample_minibatch(self):
        er = ExperienceReplay(max_size=3)
        experiences = [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9 ,0, 1)]
        for e in experiences:
            er.store_transition(state=e[0], action=e[1], reward=e[2], next_state=e[3])
        self.eq(1, len(er.sample_minibatch(minibatch_size=1)))
        with patch('random.sample', side_effect=lambda lst, num: lst[len(lst)-num:]):
            minibatch = er.sample_minibatch(minibatch_size=2)
            expected = experiences[1:]
            self.eq(expected, minibatch)

    def test_dump_load(self):
        er = ExperienceReplay(max_size=2)
        experiences = [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9 ,0, 1)]
        for e in experiences:
            er.store_transition(state=e[0], action=e[1], reward=e[2], next_state=e[3])
        dump = er.dump()

        new_er = ExperienceReplay(max_size=3)
        new_er.load(dump)
        self.eq(er.max_size, new_er.max_size)
        self.eq(er.queue, new_er.queue)


def setup_stub_task():
    mock_task = Mock()
    mock_task.generate_initial_state.return_value = 0
    mock_task.is_terminal_state.side_effect = lambda state: state == 4
    mock_task.transit_state.side_effect = lambda state, action: state + action
    mock_task.generate_possible_actions.side_effect = lambda state: [] if state == 4 else [state + 1, state + 2]
    mock_task.calculate_reward.side_effect = lambda state: state**2
    return mock_task

def gen_tmp_file_path(dir_path):
    return os.path.join(dir_path, "hoge.pickle")

def cleanup_trash():
    filenames = ["hoge.pickle", "dq_replay_memory.pickle"]
    teardown_tmp_dir(__file__, filenames)




import StringIO
import sys

from mock import Mock

from kyoka.algorithm.rl_algorithm import BaseRLAlgorithm, generate_episode
from kyoka.policy import GreedyPolicy, EpsilonGreedyPolicy
from kyoka.value_function import BaseActionValueFunction
from kyoka.callback import BaseFinishRule
from tests.base_unittest import BaseUnitTest


class BaseRLAlgorithmTest(BaseUnitTest):

    def setUp(self):
        self.algo = BaseRLAlgorithm()
        self.capture = StringIO.StringIO()
        sys.stdout = self.capture

    def tearDown(self):
        sys.stdout = sys.__stdout__

    def test_setup(self):
        algo = BaseRLAlgorithm()
        mock_func = Mock()
        algo.setup(task=0, policy=1, value_function=mock_func)
        self.eq(0, algo.task)
        self.eq(1, algo.policy)
        self.eq(mock_func, algo.value_function)
        mock_func.setup.assert_called()

    def test_save(self):
        algo = self.TestImplementation()
        mock_func = Mock()
        algo.setup(task=0, policy=1, value_function=mock_func)
        algo.save("hoge")
        mock_func.save.assert_called_with("hoge")
        self.eq(1, algo.save_count)
        self.eq(0, algo.load_count)

    def test_load(self):
        algo = self.TestImplementation()
        mock_func = Mock()
        algo.setup(task=0, policy=1, value_function=mock_func)
        algo.load("hoge")
        mock_func.load.assert_called_with("hoge")
        self.eq(0, algo.save_count)
        self.eq(1, algo.load_count)


    def test_error_msg_when_not_implement_abstract_method(self):
        self.__check_err_msg(lambda : self.algo.run_gpi_for_an_episode("dummy", "dummy", "dummy"), "run_gpi_for_an_episode")

    def test_generate_episode(self):
        task = self.__setup_stub_task()
        policy = GreedyPolicy()
        value_func = self.__setup_stub_value_function()
        episode = generate_episode(task, policy, value_func)
        self.eq(3, len(episode))
        self.eq((0, 1, 1, 1), episode[0])
        self.eq((1, 2, 3, 9), episode[1])
        self.eq((3, 4, 7, 49), episode[2])

    def test_GPI(self):
        algo = self.TestImplementation()
        task = self.__setup_stub_task()
        policy = GreedyPolicy()
        value_func = self.__setup_stub_value_function()
        algo.setup(task, policy, value_func)
        finish_rule = self.TestFinishRule()
        algo.run_gpi(nb_iteration=3, callbacks=finish_rule)
        expected = "[test_tag] finish:2\n"
        self.include(expected, self.capture.getvalue())

    def test_GPI_with_multiple_finish_rules(self):
        algo = self.TestImplementation()
        task = self.__setup_stub_task()
        policy = GreedyPolicy()
        value_func = self.__setup_stub_value_function()
        algo.setup(task, policy, value_func)
        finish_rule1 = self.TestFinishRule([False, True])
        finish_rule2 = self.TestFinishRule([True, False])
        finish_rules = [finish_rule1, finish_rule2]
        finish_msg = algo.run_gpi(nb_iteration=2, callbacks=finish_rules)
        expected = 1
        expected = "[test_tag] finish:1\n"
        self.include(expected, self.capture.getvalue())

    def test_verbose_mode(self):
        algo = self.TestImplementation()
        task = self.__setup_stub_task()
        policy = GreedyPolicy()
        value_func = self.__setup_stub_value_function()
        algo.setup(task, policy, value_func)

        capture = StringIO.StringIO()
        sys.stdout = capture
        algo.run_gpi(nb_iteration=2, callbacks=self.TestFinishRule())
        self.include("[Progress] Finished", capture.getvalue())

        capture = StringIO.StringIO()
        sys.stdout = capture
        algo.run_gpi(nb_iteration=2, callbacks=self.TestFinishRule(), verbose=0)
        self.not_include("[Progress] Finished", capture.getvalue())

    def test_defult_finish_message_must_be_logged(self):
        algo = self.TestImplementation()
        task = self.__setup_stub_task()
        policy = GreedyPolicy()
        value_func = self.__setup_stub_value_function()
        algo.setup(task, policy, value_func)
        finish_rule = self.TestFinishRule()
        algo.run_gpi(nb_iteration=2, callbacks=finish_rule)

        expected = "[Progress] Completed"
        self.include(expected, self.capture.getvalue())

    def test_not_to_log_duplicated_default_finish_message(self):
        algo = self.TestImplementation()
        task = self.__setup_stub_task()
        policy = GreedyPolicy()
        value_func = self.__setup_stub_value_function()
        algo.setup(task, policy, value_func)
        algo.run_gpi(nb_iteration=2)
        self.eq(1, self.capture.getvalue().count("[Progress] Completed"))

    def test_default_annealing_message(self):
        algo = self.TestImplementation()
        task = self.__setup_stub_task()
        policy = EpsilonGreedyPolicy()
        value_func = self.__setup_stub_value_function()

        algo.setup(task, policy, value_func)
        algo.run_gpi(nb_iteration=2)
        self.not_include("[EpsilonGreedyAnnealing]", self.capture.getvalue())

        policy.set_eps_annealing(1.0, 0.1, 100)
        algo.setup(task, policy, value_func)
        algo.run_gpi(nb_iteration=2)
        self.include("[EpsilonGreedyAnnealing]", self.capture.getvalue())

    def test_set_callback(self):
        algo = self.TestImplementation()
        value_func = Mock(name="value_func")
        algo.setup("task", "dummy", value_func)
        callback = Mock()
        callback.interrupt_gpi.return_value = False
        finish_rule = self.TestFinishRule()
        finish_msg = algo.run_gpi(nb_iteration=2, callbacks=[callback, finish_rule])
        self.eq(1, callback.before_gpi_start.call_count)
        self.eq(2, callback.before_update.call_count)
        self.eq(2, callback.after_update.call_count)
        self.eq(1, callback.after_gpi_finish.call_count)
        callback.before_gpi_start.assert_called_with("task", value_func)
        callback.before_update.assert_called_with(2, "task", value_func)
        callback.after_update.assert_called_with(2, "task", value_func)
        callback.after_gpi_finish.assert_called_with("task", value_func)

    def test_error_when_run_gpi_called_without_setup(self):
        algo = self.TestImplementation()
        with self.assertRaises(Exception) as e: algo.run_gpi(nb_iteration=2)
        self.include("setup", e.exception.message)
        self.include("run_gpi", e.exception.message)

    def __setup_stub_task(self):
        mock_task = Mock()
        mock_task.generate_initial_state.return_value = 0
        mock_task.is_terminal_state.side_effect = lambda state: state == 7
        mock_task.transit_state.side_effect = lambda state, action: state + action
        mock_task.generate_possible_actions.side_effect = lambda state: [state + 1]
        mock_task.calculate_reward.side_effect = lambda state: state**2
        return mock_task

    def __setup_stub_value_function(self):
        mock_value_func = Mock(spec=BaseActionValueFunction)
        mock_value_func.predict_value.return_value = 0
        return mock_value_func

    def __check_err_msg(self, target_method, target_word):
        try:
            target_method()
        except NotImplementedError as e:
            self.include(target_word, str(e))
        else:
            self.fail("NotImplementedError does not occur")

    class TestFinishRule(BaseFinishRule):

        def __init__(self, return_value=[False, True]):
            BaseFinishRule.__init__(self)
            self.return_value = return_value
            self.return_idx = 0

        def define_log_tag(self):
            return "test_tag"

        def check_condition(self, iteration_count, _task, _value_function):
            self.return_idx += 1
            return self.return_value[self.return_idx-1]

        def generate_start_message(self):
            return ""

        def generate_finish_message(self, iteration_count):
            return "%s:%s" % ("finish", iteration_count)


    class TestImplementation(BaseRLAlgorithm):

        def __init__(self):
            BaseRLAlgorithm.__init__(self)
            self.update_count = 0
            self.save_count = 0
            self.load_count = 0

        def save_algorithm_state(self, save_dir_path):
            self.save_count += 1

        def load_algorithm_state(self, load_dir_path):
            self.load_count += 1

        def run_gpi_for_an_episode(self, _task, _policy, _value_function):
            self.update_count += 1
            return self.update_count




from nose.tools import raises
from mock import Mock

from kyoka.policy import BasePolicy, GreedyPolicy, EpsilonGreedyPolicy
from tests.base_unittest import BaseUnitTest


class BasePolityTest(BaseUnitTest):

    def setUp(self):
        self.policy = BasePolicy()

    @raises(NotImplementedError)
    def test_choose_action(self):
        self.policy.choose_action("dummy", "dummy", "dummy")

class GreedyPolicyTest(BaseUnitTest):

    def test_choose_action_when_best_action_is_single(self):
        task = setup_task_stub([1,2,3])
        value_func = setup_value_function_stub([100, 50, 150])
        policy = GreedyPolicy()
        greedy_action = policy.choose_action(task, value_func, state="dummy")
        self.eq(3, greedy_action)

    def test_choose_action_when_best_action_is_multiple(self):
        task = setup_task_stub([1,2,3])
        value_func = setup_value_function_stub([100, 50, 100])
        random = self.setup_random()
        policy = GreedyPolicy(rand=random)
        greedy_action = policy.choose_action(task, value_func, state="dummy")
        self.eq(3, greedy_action)

    def setup_random(self):
        random = Mock()
        random.choice.side_effect = lambda ary: ary[1]
        return random

class EpsilonGreedyPolicyTest(BaseUnitTest):

    def test_choose_action_boundary_test(self):
        eps = 0.0001
        self.eq(1, self.choose_action_with_rand_value(0))
        self.eq(1, self.choose_action_with_rand_value(0.1 - eps))
        self.eq(2, self.choose_action_with_rand_value(0.1))
        self.eq(2, self.choose_action_with_rand_value(0.2 + (1-0.3) - eps))
        self.eq(3, self.choose_action_with_rand_value(0.2 + (1-0.3)))
        self.eq(3, self.choose_action_with_rand_value(1-eps))

    def test_epsilon_annealing(self):
        policy = EpsilonGreedyPolicy(eps=0.5)
        self.false(policy.do_annealing)
        policy.set_eps_annealing(initial_eps=1.0, final_eps=0.1, anneal_duration=9)
        self.true(policy.do_annealing)
        expected_eps = [(i+1)*0.1 for i in range(9)][::-1] + [0.1, 0.1]
        for eps in expected_eps:
            policy.anneal_eps()
            self.almosteq(eps, policy.eps, 0.000001)

    def xtest_sampling_action(self):
        task = setup_task_stub([1,2,3])
        value_func = setup_value_function_stub([100, 150, 50]*100000)
        policy = EpsilonGreedyPolicy(eps=0.3)
        result = [policy.choose_action(task, value_func, state="dummy") for _ in range(100000)]
        count = (result.count(1), result.count(2), result.count(3))
        self.debug()
        # sample result => count = (10062, 79955, 9983)


    def choose_action_with_rand_value(self, rand_val):
        task = setup_task_stub([1,2,3])
        value_func = setup_value_function_stub([100, 150, 50])
        random = self.setup_random(rand_val)
        policy = EpsilonGreedyPolicy(eps=0.3, rand=random)
        return policy.choose_action(task, value_func, state="dummy")

    def setup_random(self, rand_val):
        random = Mock()
        random.random.return_value = rand_val
        random.choice.side_effect = lambda ary: ary[0]
        return random



def setup_task_stub(possible_actions):
    mock_task = Mock()
    mock_task.generate_possible_actions.return_value = possible_actions
    return mock_task

def setup_value_function_stub(mock_return):
    mock_value_func = Mock()
    mock_value_func.predict_value.side_effect = mock_return
    return mock_value_func




import os

from kyoka.utils import pickle_data, unpickle_data, value_function_check
from kyoka.value_function import BaseTabularActionValueFunction, BaseApproxActionValueFunction
from kyoka.algorithm.rl_algorithm import BaseRLAlgorithm, generate_episode


class MonteCarlo(BaseRLAlgorithm):
    """Every-visit MonteCarlo method with supporting reward discounting,

    "Every-visit" means "using every state for update in an episode even if
    same state appeared in the episode".

    Algorithm is implemented based on the book "Reinforcement Learning: An Introduction"
    (reference : https://webdocs.cs.ualberta.ca/~sutton/book/bookdraft2016sep.pdf)

    - Algorithm -
    Initialize:
        T  <- your RL task
        PI <- Policy used to generate episode
        Q  <- action value function
    Repeat until computational budge runs out:
        generate an episode of T by following policy PI
        for each state-action pair (S, A)  appeared in the episode:
            G <- sum of rewards gained after state S (discounted if gamma < 1)
            Q(S, A) <- average G of S sampled ever
    """

    def __init__(self, gamma=1):
        """
        If you want to discount future reward then set gamma < 1.

        For example, we have an apisode like this
        episode :
            (state0, action0) -> reward0 ->
            (state1, action1) -> reward1 ->
            (state2, action2) -> reward2 -> finish

        then reward discounting is done like this
            reward_sum_from_state0 = reward0 + gamma * reward1 + gamma**2 reward2

        Args:
            gamma : discount factor of reward. default=1. 0 < gamma <= 1.
         """
        self.gamma = gamma

    def setup(self, task, policy, value_function):
        validate_value_function(value_function)
        super(MonteCarlo, self).setup(task, policy, value_function)

    def run_gpi_for_an_episode(self, task, policy, value_function):
        episode = generate_episode(task, policy, value_function)
        for idx, turn_info in enumerate(episode):
            state, action, _next_state, _reward = turn_info
            following_reward = self._calculate_following_state_reward(idx, episode)
            value_function.backup(state, action, following_reward, alpha="dummy")

    def _calculate_following_state_reward(self, current_turn, episode):
        following_turn_info = episode[current_turn:]
        following_reward = [reward for _, _, _, reward in following_turn_info]
        return sum([self.__discount(step, reward) for step, reward in enumerate(following_reward)])

    def __discount(self, step, reward):
        return self.gamma ** step * reward

class MonteCarloTabularActionValueFunction(BaseTabularActionValueFunction):
    """Tabular action value function for MonteCarlo method.

    Backup target passed from MonteCarlo is G(reward sum of state S).
    So backup is done just averaging G of S sampled ever.

    Calculation of average is implemented by memory efficient way.
    ("_calc_average_in_incremental_way" is the method calculates average")
    """

    SAVE_FILE_NAME = "montecarlo_update_counter.pickle"

    def setup(self):
        super(MonteCarloTabularActionValueFunction, self).setup()
        self.update_counter = self.generate_initial_table()

    def define_save_file_prefix(self):
        return "montecarlo"

    def save(self, save_dir_path):
        super(MonteCarloTabularActionValueFunction, self).save(save_dir_path)
        pickle_data(self._gen_update_counter_file_path(save_dir_path), self.update_counter)

    def load(self, load_dir_path):
        super(MonteCarloTabularActionValueFunction, self).load(load_dir_path)
        if not os.path.exists(self._gen_update_counter_file_path(load_dir_path)):
            raise IOError(
                    'The saved data of "MonteCarlo" algorithm is not found in [ %s ]' %
                    load_dir_path)
        self.update_counter = unpickle_data(self._gen_update_counter_file_path(load_dir_path))

    def backup(self, state, action, backup_target, alpha):
        update_count = self.fetch_value_from_table(self.update_counter, state, action)
        Q_value = self.fetch_value_from_table(self.table, state, action)
        new_value = self._calc_average_in_incremental_way(update_count, backup_target, Q_value)
        self.insert_value_into_table(self.table, state, action, new_value)
        self.insert_value_into_table(self.update_counter, state, action, update_count+1)

    def _calc_average_in_incremental_way(self, k, r, Q):
        """Memory efficient implementation to calculate average"""
        return Q + 1.0 / (k + 1) * (r - Q)

    def _gen_update_counter_file_path(self, dir_path):
        return os.path.join(dir_path, self.SAVE_FILE_NAME)

class MonteCarloApproxActionValueFunction(BaseApproxActionValueFunction):
    """Approximation action value function for MonteCarlo method.
    There is no additional method from base class to use MonteCarlo method.

    Backup target passed from MonteCarlo is G(reward sum of state S).
    So backup should be done to approximate average of G of S sampled ever.
    """
    pass

def validate_value_function(value_function):
    value_function_check("MonteCarlo",
            [MonteCarloTabularActionValueFunction, MonteCarloApproxActionValueFunction],
            value_function)




import os
import random

from kyoka.utils import pickle_data, unpickle_data, value_function_check, build_not_implemented_msg
from kyoka.policy import GreedyPolicy, EpsilonGreedyPolicy
from kyoka.value_function import BaseApproxActionValueFunction
from kyoka.algorithm.rl_algorithm import BaseRLAlgorithm, generate_episode


class DeepQLearning(BaseRLAlgorithm):
    """deep Q-learning method using experience replay proposed in
    the paper "Human-level control through deep reinforcement learning".

    deep Q-learning arranges Q-Learning in 2 points to improve and
    get stability of learing process.
    - Use mini-batch of experience to update value function.
    - Separate value function used to "create backup target".

    Algorithm is implemented based on the original paper.
    (reference: http://www.nature.com/nature/journal/v518/n7540/abs/nature14236.html)

    - Algorithm -
    Parameter
        g <- discounting factor (gamma)
        N <- capacity of replay memory
        C <- interval to sync Q' with Q
        minibatch_size <- size of minibatch used to train Q
        replay_start_size <- initial size of replay memory. Fill D
            with this number of experiences which created by random policy.
            This procedure is done in setup phase.

    Initialize
        T  <- your RL task
        PI <- policy used to "select action during the episode"
        Q  <- approximate action value function (ex. use neural network)
        Q' <- separated value function created by deepcopy Q
        D <- replay memory of capacity N filled with experience of initial size
    Repeat until computational budget runs out:
        S <- generate initial state of task T
        A <- choose action at S by following PI
        Repeat until S is terminal state:
            S' <- next state of S after taking action A
            R <- reward gained by taking action A at state S
            A' <- next action at S' by following policy PI
            append experience (S,A,R,S') to D
            MB <- sample "minibatch_size" experiences from D
            BT <- transform minibatch of experiences into backup targets
            (BT = [r + g * Q(s', GA) for s,a,r,s' in MB], GA=greedy action at s')
            Update Q by using BT (minibatch of backup targets)
            Every C step: Q' <- Q (ex. deepcopy weights of Q to Q')
            S, A <- S', A'
    """

    SAVE_FILE_NAME = "dq_replay_memory.pickle"

    def __init__(self, gamma=0.99, N=1000000, C=10000,
            minibatch_size=32, replay_start_size=50000):
        """
        Args:
            g <- discounting factor (gamma)
            N <- capacity of replay memory
            C <- interval to sync Q' with Q
            minibatch_size <- size of minibatch used to train Q
            replay_start_size <- initial size of replay memory.
        """

        self.gamma = gamma
        self.replay_memory = ExperienceReplay(max_size=N)
        self.C = C
        self.minibatch_size = minibatch_size
        self.replay_start_size = replay_start_size
        self.reset_step_counter = 0

    def setup(self, task, policy, value_function):
        validate_value_function(value_function)
        super(DeepQLearning, self).setup(task, policy, value_function)
        initialize_replay_memory(task, value_function, self.replay_memory, self.replay_start_size)
        self.greedy_policy = GreedyPolicy()

    def run_gpi_for_an_episode(self, task, policy, value_function):
        value_function.use_target_network(False)
        state = task.generate_initial_state()

        while not task.is_terminal_state(state):
            action = policy.choose_action(task, value_function, state)
            next_state = task.transit_state(state, action)
            reward = task.calculate_reward(next_state)
            self.replay_memory.store_transition(state, action, reward, next_state)
            state = next_state

            experience_minibatch = self.replay_memory.sample_minibatch(self.minibatch_size)
            backup_minibatch = self._gen_backup_minibatch(
                    task, self.greedy_policy, value_function, experience_minibatch)
            value_function.backup_on_minibatch(value_function.q_network, backup_minibatch)

            if self.reset_step_counter >= self.C:
                value_function.reset_target_network()
                self.reset_step_counter = 0
            else:
                self.reset_step_counter += 1

    def save_algorithm_state(self, save_dir_path):
        """Save initial params, replay memory and counter for sync Q' with Q"""
        state = (
                self.gamma, self.replay_memory.dump(), self.C, self.minibatch_size,
                self.replay_start_size, self.reset_step_counter
                )
        pickle_data(self._gen_replay_memory_save_path(save_dir_path), state)

    def load_algorithm_state(self, load_dir_path):
        """Load initial params, replay memory and counter for sync Q' with Q"""
        state = unpickle_data(self._gen_replay_memory_save_path(load_dir_path))
        (self.gamma, replay_memory_serial, self.C, self.minibatch_size,
                self.replay_start_size, self.reset_step_counter) = state
        self.greedy_policy = GreedyPolicy()
        new_replay_memory = ExperienceReplay(max_size=1)
        new_replay_memory.load(replay_memory_serial)
        self.replay_memory = new_replay_memory

    def _gen_backup_minibatch(self, task, greedy_policy, value_function, experience_minibatch):
        """Create minibatch of backup targets from minibatch of experiences
        Returns
            backup_minibatch : minibatch of training data for value function.
                               It's array of learning data which is tuple of
                               (state, action, backup_target).
                               Most of the case value function is trained by
                               using MSE between Q(state, action) and backup_target.
        """
        value_function.use_target_network(True)
        backup_minibatch = [
                self._gen_backup_data(task, greedy_policy, value_function, experience)
                for experience in experience_minibatch]
        value_function.use_target_network(False)
        return backup_minibatch

    def _gen_backup_data(self, task, greedy_policy, value_function, experience):
        """Transform experience into backup targets in training data format
        Returns
            learning data: tuple of (state, action, backup_target).
                           value function receives minibatch of this tuples and
                           train value function maybe like below.
                           MSE <- ( Q(state, action) - backup_target )^2
        """
        state, action, reward, next_state = experience
        greedy_action = choose_action(task, greedy_policy, value_function, next_state)
        greedy_Q_value = predict_value(value_function, next_state, greedy_action)
        backup_target = reward + self.gamma * greedy_Q_value
        return (state, action, backup_target)

    def _gen_replay_memory_save_path(self, dir_path):
        return os.path.join(dir_path, self.SAVE_FILE_NAME)

class DeepQLearningApproxActionValueFunction(BaseApproxActionValueFunction):
    """Base class of Approximation value function for deep Q-learing.

    Child class required to implement following 6 methods.
    - initialize_network
    - deepcopy_network
    - predict_value_by_network
    - backup_on_minibatch
    - save_networks (if you want save/load feature)
    - load_networks (if you want save/load feature)
    """

    def initialize_network(self):
        """Initialize action value function Q
        Returns:
            q_network: Item returned here is used as Q value function.
                       Most of the case, you initialize neuralnetwork here
                       and return it.
                       You can access this item by "value_function.q_network".
        """
        err_msg = build_not_implemented_msg(self, "initialize_network")
        raise NotImplementedError(err_msg)

    def deepcopy_network(self, q_network):
        """Deepcopy Q to initialize and sync separated network Q'
        Args:
            q_network: Q value function initialized by "initialize_network"
        Returns:
            q_hat_network: deepcopy of q_network which passed as argument.
        """
        err_msg = build_not_implemented_msg(self, "deepcopy_network")
        raise NotImplementedError(err_msg)

    def predict_value_by_network(self, network, state, action):
        """Define how to predict action value by Q and Q' which you defined
        Args:
            network: q_network or q_hat_network which initialized by
                     "initialize_network" or "deepcopy_network"
            state: state of state-action pair to predict value
            action: action of state-action pair to predict value
        Returns:
            predicted_value: prediction value of action by passed network
        """
        err_msg = build_not_implemented_msg(self, "predict_value_by_network")
        raise NotImplementedError(err_msg)

    def backup_on_minibatch(self, q_network, backup_minibatch):
        """Define how to train Q network which you defined
        Args:
            q_network: Q value function initialized by "initialize_network"
            backup_minibatch : minibatch of training data for Q. It's array of
                               learning data which is tuple of
                               (state, action, backup_target).
                               Most of the case value function is trained by
                               using MSE between Q(state, action) and backup_target.
        """
        err_msg = build_not_implemented_msg(self, "backup_on_minibatch")
        raise NotImplementedError(err_msg)

    def save_networks(self, q_network, q_hat_network, save_dir_path):
        """Save Q and Q' under passed directory
        Args:
            q_network: Q value function initialized by "initialize_network"
            q_hat_network: Q' deepcopied by "deepcopy_network"
            save_dir_path: path to directory to save. Load method would be
                           called on this directory.
        """
        err_msg = build_not_implemented_msg(self, "save_networks")
        raise NotImplementedError(err_msg)

    def load_networks(self, load_dir_path):
        """Load Q and Q' from passed directory
        Args:
            load_dir_path: path to directory which you executed "save_networks"
        Returns:
            networks: tuple of Q and Q' which loaded from passed directory.
                      Return loaded networks like below.

                      q_network = # load some way
                      q_hat_network = # load some way
                      return q_network, q_hat_network
        """
        err_msg = build_not_implemented_msg(self, "load_networks")
        raise NotImplementedError(err_msg)


    def setup(self):
        """Initialize Q and Q'"""
        self.q_network = self.initialize_network()
        self.reset_target_network()
        self.use_target_network_flg = False

    def use_target_network(self, use_target_network):
        """Switch value function for prediction between Q and Q'
        Use Q on prediction if use_target_network=True else Q'
        """
        self.use_target_network_flg = use_target_network

    def predict_value(self, state, action):
        """Predict value of state-action pair with Q or Q'
        Main logic of prediction is delegated to abstract method
        "predict_value_by_network".

        Switch value function by watching self.use_target_network_flg.
        You can switch this flg by using "use_target_network_flg(True/False)".
        """
        network = self.q_hat_network if self.use_target_network_flg else self.q_network
        return self.predict_value_by_network(network, state, action)

    def reset_target_network(self):
        """Sync Q' with Q by calling user-defined method "deepcopy_network" """
        self.q_hat_network = self.deepcopy_network(self.q_network)

    def save(self, save_dir_path):
        """Save networks under passed directory. Main logic is delegated to
        abstract method "save_networks".
        """
        self.save_networks(self.q_network, self.q_hat_network, save_dir_path)

    def load(self, load_dir_path):
        """Load networks from passed directory and set them as property. Main
        logic is delegated to abstract method "load_networks".
        """
        self.q_network, self.q_hat_network = self.load_networks(load_dir_path)


class ExperienceReplay(object):
    """Implementation of ExperienceReplayMemory"""

    def __init__(self, max_size):
        """
        Args:
            max_size: capacity of replay memory. If size of memory exceeds
                      after store_transition, old item will be poped from
                      memory. (FIFO)
        """
        self.max_size = max_size
        self.queue = []

    def store_transition(self, state, action, reward, next_state):
        """Store new experience and pop old item if it reaches max size"""
        if len(self.queue) >= self.max_size:
            self.queue.pop(0)
        self.queue.append((state, action, reward, next_state))

    def sample_minibatch(self, minibatch_size):
        """Return array of experiences with specified size by sampling
        from replay memory at random.
        """
        return random.sample(self.queue, minibatch_size)

    def dump(self):
        return (self.max_size, self.queue)

    def load(self, serial):
        self.max_size, self.queue = serial

def initialize_replay_memory(task, value_function, replay_memory, start_size):
    """Fill passed replay memory with specified size of experience. Experience
    is created by generating episode with random policy until expected size
    of experience is corrected.
    """
    random_policy = EpsilonGreedyPolicy(eps=1.0)
    while len(replay_memory.queue) < start_size:
        episode = generate_episode(task, random_policy, value_function)
        for state, action, next_state, reward in episode:
            replay_memory.store_transition(state, action, reward, next_state)
            if len(replay_memory.queue) >= start_size: return

ACTION_ON_TERMINAL_FLG = "action_on_terminal"

def choose_action(task, policy, value_function, state):
    if task.is_terminal_state(state):
        return ACTION_ON_TERMINAL_FLG
    else:
        return policy.choose_action(task, value_function, state)

def predict_value(value_function, next_state, next_action):
    if ACTION_ON_TERMINAL_FLG == next_action:
        return 0
    else:
        return value_function.predict_value(next_state, next_action)

def validate_value_function(value_function):
    value_function_check("DeepQLearning",
            [DeepQLearningApproxActionValueFunction],
            value_function)




from setuptools import setup, find_packages

setup(
    name = "kyoka",
    version = "0.2.1",
    author = "ishikota",
    author_email = "ishikota086@gmail.com",
    description = ("Simple Reinforcement Learning Library"),
    license = "MIT",
    keywords = "reinforcement learning RL",
    url = "https://github.com/ishikota/kyoka",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
    ],
)



import random
import math

from kyoka.utils import build_not_implemented_msg
from kyoka.task import BaseTask


class BaseMCTS(object):
    """Base class to implement MonteCarloTreeSearch method.

    This class implements base logic of MonteCarloTreeSearch algorithm like
    building and traversing game tree, backpropagation of simulation result.
    (These base procedure is implemented in "planning" method)

    But "how to build node of game tree from state" and "how to calculate
    value of edge(action)" is abstracted. (Because these points are strongly
    depend on the task to solve.)
    So you need to implement these points by creating your own
    Node(child class of BaseNode) and Edge(child class of BaseEdge) classes.
    And return that Node in "generate_node_from_state" method of this class.

    We prepared implmentation of UCT search as UCTNode class. If you want to
    use UCT search, return UCTNode in "generate_node_from_state" by
    initializing node with passed state like below.

        def generate_node_from_state(self, state):
            return UCTNode(self.task, state)

    If you want to use algorithm not UCT search, please implement your Node and
    Edge class referring the imeplemntation of UCTNode and UCTEdge classes.

    Algorithm is implemented based on the papaer
    "A Survey of Monte Carlo Tree Search Methods"
    (reference: http://ieeexplore.ieee.org/document/6145622/)
    """

    def __init__(self, task):
        """
        Args:
            task: task object to apply MCTS method
        """
        self.task = task
        self.playout_policy = random_playout
        self.last_calculated_tree = None
        self.finish_rule = None

    def generate_node_from_state(self, state):
        """Transform state of the task into Node(child class of BaseNode).

        Args:
            state: you return node of tree which represents this state
        Returns:
            node: node which represents passed state. Node must be child of
                  BaseNode class in this module.
        """
        err_msg = build_not_implemented_msg(self, "generate_node_from_state")
        raise NotImplementedError(err_msg)

    def set_playout_policy(self, policy):
        """Define how to play simulation(playout). Default is random play.

        In default, "random_playout" method is set as playout policy.
        If you want to update this behavior create your own playout policy
        and set it by this method.
        The format of playout policy method to pass this method would be
        like this.

            def my_playout_policy(self, task, leaf_node):
                current_state = leaf_node.state
                while not task.is_terminal_state(state):
                    state = current_node.state
                    action = choose_action_by_my_logic(state)
                    state = task.transit_state(state, action)
                return task.calculate_reward(state)

        And set above method like this.
            my_mcts.set_playout_policy(my_playout_policy)

        Args:
            policy: the method to run a simulation. this is used in _playout
                    method of BaseMCTS class.
        """
        self.playout_policy = policy

    def set_finish_rule(self, finish_rule):
        """Set finish rule of planning if you want to call "planning" method
        through "choose_action" method.
        """
        self.finish_rule = finish_rule

    def choose_action(self, _task, _value_function, state):
        """Utility method for calling "planning" method in the common format of
        this library. If you use this method, you must call "set_finish_rule"
        before calling this method.

        Raises:
            AssertionError: if "set_finish_rule" is not called
        """
        assert self.finish_rule is not None
        return self.planning(state, self.finish_rule)

    def planning(self, state, finish_rule):
        """Main logic of MCTS.

        - Algorithm -
        Initialize:
            R <- Build game tree which has only a root node which represents
                 passed state
        Repeat until satisfies finish condition of passed finish_rule:
            N <- descend R until not expanded node found.
                 (expanded = visited all child edge at least once)
            N' <- child node of new edge expanded at N
            R <- reward of simulation started from the state which N' represents
            Backpropagate R from N' to R
        return best action(edge of highest value) at R

        Args:
            state: start state to run MCTS procedure. It would be current state
                   of your agent in the task.
            finish_rule: child class of "kyoka.callback.FinishRule". This rule
                         is used to determine when to finish the MCTS iteration.
        Returns:
            action: best action at passed state calculated by MCTS algorithm.
        Raises:
            AssertionError: if passed state is terminal state, nothing to do.
        """
        assert not self.task.is_terminal_state(state)
        _log_start_msg(finish_rule)
        iteration_count = 0

        # TODO use last calculation result like AlphaGo
        root_node = self.generate_node_from_state(state)
        while not finish_rule.check_condition(iteration_count, self.task, None):
            finish_rule.before_update(iteration_count, self.task, None)

            selected_node = self._select(root_node)
            if self.task.is_terminal_state(selected_node.state):
                leaf_node = selected_node
                reward = self.task.calculate_reward(leaf_node.state)
            else:
                leaf_node = self._expand(selected_node)
                reward = self._playout(leaf_node)
            self._backpropagation(leaf_node, reward)

            finish_rule.after_update(iteration_count, self.task, None)
            iteration_count += 1

        self.last_calculated_tree = root_node
        _log_finish_msg(finish_rule, iteration_count)
        return root_node.greedy_edge.action

    def _select(self, root_node):
        """Find terminal or not expanded node"""
        target_node = root_node
        while target_node.is_expanded:
            target_node = target_node.select_best_edge().child_node
        return target_node

    def _expand(self, node):
        """Expand new edge from passed node"""
        assert node.has_unvisited_edge
        edge = node.select_unvisited_edge()
        edge.build_child(self.generate_node_from_state)
        return edge.child_node

    def _playout(self, leaf_node):
        """Run simulation from state of passed node and return reward"""
        reward = self.playout_policy(self.task, leaf_node)
        return reward

    def _backpropagation(self, leaf_node, reward):
        """Backpropagate passed reward from leaf_node to root node"""
        target = leaf_node
        assert target.parent_edge is not None
        while target.parent_edge:
            target.parent_edge.visit()
            target.parent_edge.update_by_new_reward(reward)
            target = target.parent_edge.parent_node

def random_playout(task, leaf_node, rand=random):
    """Default implmentation of playout policy of BaseMCTS.
    Simulation is played by the action choosed from possible actions
    at random.

    Args:
        task: run simulation by using this task
        leaf_node: start simulation from leaf_node.state
        rand: rand.choice is used to choose action
    Returns:
        reward: result of simulation
    """
    state = leaf_node.state
    while not task.is_terminal_state(state):
        actions = task.generate_possible_actions(state)
        action = random.choice(actions)
        state = task.transit_state(state, action)
    return task.calculate_reward(state)

def _log_start_msg(finish_rule):
    finish_rule.log(finish_rule.generate_start_message().replace("GPI", "MCTS"))

def _log_finish_msg(finish_rule, iteration_count):
    fin_msg= finish_rule.generate_finish_message(iteration_count)
    finish_rule.log(fin_msg.replace("GPI", "MCTS"))

class BaseNode(object):
    """Base class to build Node class of tree for MCTS

    This class is responsible for the logic of
    1. "how to build an Edge from the action"
    2. "how to choose the child node to descent in Select procedure of MCTS"

    The logic of 1. is abstracted as "generate_edge" method.
    The logic of 2. is implemented as "select_best_edge" method. Default
    implementation choose the edge which has highest value.

    Property:
        task: task passed from "BaseMCTS.generate_node_from_state"
        state: state of the node passed from "BaseMCTS.generate_node_from_state"
        parent_edge: reference of parent edge to access parent node
        child_edges: array of reference of edge to access child nodes
        greedy_edge: the edge which has highest average_reward
        terminal: boolean. True if state of this node is terminal.
        is_expanded: boolean. False if this node is possible to expand.
        has_unvisited_edge: boolean. True if one of edge.visit_count is 0.
        visit_count: sum of visit counts of child edges
    """

    def __init__(self, task, state):
        self.task = task
        self.state = state
        self.parent_edge = None
        self.child_edges = self.build_child_edges(task, state)
        self.terminal = task.is_terminal_state(state)

    def generate_edge(self, parent_node, action):
        """Define how to create Edge object from passed action

        If you have MyEdge class and use it for all child edges, code is simple

            def generate_edge(self, parent_node, action):
                return MyEdge(parent_node, action)

        Args:
            parent_node: parent node of the edge you will create
            action: you create the Edge object which represents this action
        Returns:
            edge: Edge object(child class of BaseEdge) which represents the
                  passed action
        """
        err_msg = build_not_implemented_msg(self, "generate_edge")
        raise NotImplementedError(err_msg)

    def select_best_edge(self):
        """Define how to choose next node to descend in SELECT procedure"""
        return sorted([edge for edge in self.child_edges], key=self._edge_sort_key)[-1]

    def _edge_sort_key(self, edge):
        return edge.calculate_value()

    def build_child_edges(self, task, state):
        actions = task.generate_possible_actions(state)
        return [self.generate_edge(self, action) for action in actions]

    def select_unvisited_edge(self):
        return [edge for edge in self.child_edges if not edge.has_child][0]

    @property
    def greedy_edge(self):
        best = max([edge.average_reward for edge in self.child_edges])
        best_edges = [edge for edge in self.child_edges if edge.average_reward==best]
        return best_edges[0]

    @property
    def is_expanded(self):
        """boolean. False if this node is possible to expand."""
        return not (self.terminal or self.has_unvisited_edge)

    @property
    def has_unvisited_edge(self):
        """boolean. True if one of edge.visit_count is 0."""
        return any([not edge.has_child for edge in self.child_edges])

    @property
    def visit_count(self):
        """sum of visit counts of child edges"""
        return sum([edge.visit_count for edge in self.child_edges])

class BaseEdge(object):
    """Base class to build Edge class of tree for MCTS

    This class is responsible for the logic of
    1. "holds average of rewards received in backpropagation"
    2. "how to calculate the value of edge for Select of MCTS"

    There are lots of methods to calculate value for selection 
    like UCT, UCT+, ... . So this logic is abstracted as
    "calculate_value" method.
    When you implement "calculate_value" by your own, we recommend you
    to refer implementation of UCTEdge class.

    Property:
        parent_node: reference to the parent node of the edge
        child_node: reference to the child node. None until
                    "build_child" is called.
        visit_count: the number of update in backpropagation process
        has_child: True if already "build_child" is called. This means
                   "this node is already selected at least once".
        average_reward: average of rewards received ever
    """

    def __init__(self, parent_node, action):
        self.action = action
        self.parent_node = parent_node
        self.child_node = None
        self.visit_count = 0
        self.average_reward = 0

    def calculate_value(self):
        """Define how to calculate the value of this Edge.
        This is called from select procedure of MCTS.

        Returns:
            edge_value: the value of edge used in selection procedure.
        """
        err_msg = build_not_implemented_msg(self, "calculate_value")
        raise NotImplementedError(err_msg)

    def build_child(self, state2node):
        child_state = self.parent_node.task.transit_state(self.parent_node.state, self.action)
        self.child_node = state2node(child_state)
        self.child_node.parent_edge = self

    def visit(self):
        self.visit_count += 1

    def update_by_new_reward(self, new_reward):
        self.average_reward = self._calc_average_in_incremental_way(
                self.average_reward, self.visit_count, new_reward)

    def _calc_average_in_incremental_way(self, old_value, visit_count, new_reward):
        assert visit_count != 0
        return old_value + 1.0 / visit_count * (new_reward - old_value)

    @property
    def has_child(self):
        """True if already "build_child" is called. This means
        "this node is already selected at least once".
        """
        return self.child_node is not None


class UCTNode(BaseNode):
    """Concrete implementation of BaseNode for UCT search.
    All of the logic of UCT search is implemented in UCT edge class
    """

    def generate_edge(self, parent_node, action):
        return UCTEdge(parent_node, action)

class UCTEdge(BaseEdge):
    """Concrete implementation of BaseEdge for UCT search.

    The value of edge for selection is calculated by following equation

    UCT = average(past_rewards) + 2 * C * sqrt( 2 * log(N) / n )
        where C = hyper parameter, n = visit count of edge,
              N = visit count of parent Node

    The default value of C is "1 / sqrt(2)". Because if reward is in range of
    [0,1], this value of UCT search satisfies Hoeffding's inequality.
    """

    def __init__(self, parent_node, action):
        super(UCTEdge, self).__init__(parent_node, action)
        self.C = 0.7071067811865475 # 1 / math.sqrt(2)

    def calculate_value(self):
        if self.visit_count == 0:
            explore_term = float('inf')
        else:
            explore_term = math.sqrt(2 * math.log(self.parent_node.visit_count) / self.visit_count)
        return self.average_reward + 2 * self.C * explore_term




import os
import sys
import StringIO

from nose.tools import raises
from mock import patch, Mock

from kyoka.callback import BaseCallback, BasePerformanceWatcher, EpsilonAnnealer,\
        LearningRecorder, BaseFinishRule, ManualInterruption, WatchIterationCount
from kyoka.policy import EpsilonGreedyPolicy
from tests.base_unittest import BaseUnitTest
from tests.utils import generate_tmp_dir_path, setup_tmp_dir, teardown_tmp_dir, remove_leaf_dir


def capture_log(instance):
    instance.capture = StringIO.StringIO()
    sys.stdout = instance.capture

def release_capture():
   sys.stdout = sys.__stdout__

class BaseCallbackTest(BaseUnitTest):

    def setUp(self):
        self.callback = BaseCallback()
        capture_log(self)

    def tearDown(self):
        release_capture()

    def test_interrupt_gpi(self):
        self.false(self.callback.interrupt_gpi("dummy", "dummy", "dummy"))

    def test_define_log_tag(self):
        self.eq("BaseCallback", self.callback.define_log_tag())

    def test_log(self):
        self.callback.log("hoge")
        self.eq("[BaseCallback] hoge\n", self.capture.getvalue())

class BasePerformanceWatcherTest(BaseUnitTest):

    def setUp(self):
        capture_log(self)

    def tearDown(self):
        release_capture()

    @raises(NotImplementedError)
    def test_implementation_check_on_define_performance_test_interval(self):
        BasePerformanceWatcher().define_performance_test_interval()

    @raises(NotImplementedError)
    def test_implementation_check_on_run_performance_test(self):
        BasePerformanceWatcher().run_performance_test("dummy", "dummy")

    def test_default_log_message(self):
        watcher = self.TestMinimumImplementation()
        watcher.before_gpi_start("dummy", "dummy")
        watcher.after_update(1, "dummy", "dummy")
        watcher.after_update(2, "dummy", "dummy")
        expected = "[TestMinimumImplementation] Performance test result : 3 (nb_iteration=2)\n"
        self.eq(expected, self.capture.getvalue())

    def test_setup_is_called(self):
        watcher = self.TestCompleteImplementation()
        self.false(watcher.is_setup_called)
        watcher.before_gpi_start("dummy", "dummy")
        self.true(watcher.is_setup_called)

    def test_teardown_is_called(self):
        watcher = self.TestCompleteImplementation()
        self.false(watcher.is_teardown_called)
        watcher.after_gpi_finish("dummy", "dummy")
        self.true(watcher.is_teardown_called)

    def test_timing_of_performance_test(self):
        watcher = self.TestCompleteImplementation()
        watcher.before_gpi_start("dummy", "dummy")
        watcher.after_update(1, "dummy", "dummy")
        self.eq('', self.capture.getvalue())
        watcher.after_update(2, "dummy", "dummy")
        self.eq('[Test] test:1\n', self.capture.getvalue())
        watcher.after_update(3, "dummy", "dummy")
        self.eq('[Test] test:1\n', self.capture.getvalue())
        watcher.after_update(4, "dummy", "dummy")
        self.eq('[Test] test:1\n[Test] test:4\n', self.capture.getvalue())
        self.eq([1, 4], watcher.performance_log)

    class TestCompleteImplementation(BasePerformanceWatcher):

        def __init__(self):
            self.is_setup_called = False
            self.is_teardown_called = False
            self.test_count = 0

        def setUp(self, task, value_function):
            self.is_setup_called = True

        def tearDown(self, task, value_function):
            self.is_teardown_called = True

        def define_performance_test_interval(self):
             return 2

        def run_performance_test(self, _task, _value_function):
            self.test_count += 1
            return self.test_count**2

        def define_log_message(self, iteration_count, task, value_function, test_result):
            return "test:%s" % test_result

        def define_log_tag(self):
            return "Test"

    class TestMinimumImplementation(BasePerformanceWatcher):

        def define_performance_test_interval(self):
            return 2

        def run_performance_test(self, _task, _value_function):
            return 3

class EpsilonAnnealerTest(BaseUnitTest):

    def setUp(self):
        self.policy = EpsilonGreedyPolicy()
        self.policy.set_eps_annealing(1.0, 0.1, 100)
        self.annealer = EpsilonAnnealer(self.policy)
        capture_log(self)

    def tearDown(self):
        release_capture()

    def test_define_log_tag(self):
        self.eq("EpsilonGreedyAnnealing", self.annealer.define_log_tag())

    def test_start_log(self):
        self.annealer.before_gpi_start("dummy", "dummy")
        self.include(str(1.0), self.capture.getvalue())
        self.include(str(0.1), self.capture.getvalue())

    def test_finish_log(self):
        [self.annealer.after_update(_, "dummy", "dummy") for _ in range(99)]
        self.not_include("finish", self.capture)
        self.annealer.after_update(100, "dummy", "dummy")
        self.include("finish", self.capture.getvalue())
        self.include(str(100), self.capture.getvalue())
        # Check that finish message is logged only once
        capture = StringIO.StringIO()
        sys.stdout = capture
        self.annealer.after_update(101, "dummy", "dummy")
        self.not_include("finish", capture.getvalue())

class LearningRecorderTest(BaseUnitTest):

    def setUp(self):
        self.algo = Mock()
        self.recorder = LearningRecorder(self.algo, generate_tmp_dir_path(__file__), 2)
        capture_log(self)

    def tearDown(self):
        release_capture()
        gen_dpath = lambda fname: os.path.join(generate_tmp_dir_path(__file__), fname)
        for idx in range(1, 5):
            dname = "after_%d_iteration" % idx
            remove_leaf_dir(gen_dpath(dname), [])
        remove_leaf_dir(gen_dpath("gpi_finished"), [])
        teardown_tmp_dir(__file__, [])

    def test_start_log(self):
        setup_tmp_dir(__file__)
        self.recorder.before_gpi_start("dummy", "dummy")
        self.include(str(2), self.capture.getvalue())
        self.include(generate_tmp_dir_path(__file__), self.capture.getvalue())

    @raises(Exception)
    def test_before_gpi_start(self):
        self.recorder.before_gpi_start("dummy", "dummy")

    def test_after_update(self):
        setup_tmp_dir(__file__)
        gen_dpath = lambda fname: os.path.join(generate_tmp_dir_path(__file__), fname)
        self.recorder.after_update(1, "dummy", "dummy")
        self.algo.save.assert_not_called()
        self.false(os.path.exists(gen_dpath("after_1_iteration")))
        self.recorder.after_update(2, "dummy", "dummy")
        self.true(os.path.exists(gen_dpath("after_2_iteration")))
        self.algo.save.assert_called_with(gen_dpath("after_2_iteration"))
        self.recorder.after_update(3, "dummy", "dummy")
        self.false(os.path.exists(gen_dpath("after_3_iteration")))
        self.algo.save.assert_called_with(gen_dpath("after_2_iteration"))
        self.recorder.after_update(4, "dummy", "dummy")
        self.true(os.path.exists(gen_dpath("after_4_iteration")))
        self.algo.save.assert_called_with(gen_dpath("after_4_iteration"))
        self.include(gen_dpath("after_4_iteration"), self.capture.getvalue())

    def test_after_gpi_finish(self):
        setup_tmp_dir(__file__)
        gen_dpath = lambda fname: os.path.join(generate_tmp_dir_path(__file__), fname)
        gen_dpath = lambda fname: os.path.join(generate_tmp_dir_path(__file__), fname)
        self.false(os.path.exists(gen_dpath("gpi_finished")))
        self.recorder.after_gpi_finish("dummy", "dummy")
        self.true(os.path.exists(gen_dpath("gpi_finished")))
        self.algo.save.assert_called_with(gen_dpath("gpi_finished"))

class BaseFinishRuleTest(BaseUnitTest):

    def setUp(self):
        self.rule = BaseFinishRule()
        capture_log(self)

    def tearDown(self):
        release_capture()

    @raises(NotImplementedError)
    def test_check_condition(self):
        self.rule.check_condition("dummy", "dummy", "dummy")

    @raises(NotImplementedError)
    def test_generate_start_message(self):
        self.rule.generate_start_message()

    @raises(NotImplementedError)
    def test_generate_finish_message(self):
        self.rule.generate_finish_message("dummy")

    def test_interrupt_gpi(self):
        rule = self.TestImplementation()
        self.false(rule.interrupt_gpi(1, "dummy", "dummy"))
        self.true(rule.interrupt_gpi(2, "dummy", "dummy"))
        self.false(rule.interrupt_gpi(3, "dummy", "dummy"))

    def test_log_start_message(self):
        rule = self.TestImplementation()
        rule.before_gpi_start("dummy", "dummy")
        self.eq('[test_tag] start\n', self.capture.getvalue())

    def test_log_finish_message(self):
        rule = self.TestImplementation()
        self.false(rule.interrupt_gpi(1, "task", "value_function"))
        self.eq('', self.capture.getvalue())
        self.true(rule.interrupt_gpi(2, "task", "value_function"))
        self.eq('[test_tag] finish:2\n', self.capture.getvalue())

    class TestImplementation(BaseFinishRule):

        def define_log_tag(self):
            return "test_tag"

        def check_condition(self, iteration_count, _task, _value_function):
            return iteration_count == 2

        def generate_start_message(self):
            return "start"

        def generate_finish_message(self, iteration_count):
            return "%s:%s" % ("finish", iteration_count)

class ManualInterruptionTest(BaseUnitTest):

    def setUp(self):
        file_path = self.__generate_tmp_file_path()
        self.rule = ManualInterruption(monitor_file_path=file_path, watch_interval=5)

    def tearDown(self):
        file_path = self.__generate_tmp_file_path()
        if os.path.isfile(file_path): os.remove(file_path)

    def test_define_log_tag(self):
        self.eq("ManualInterruption", self.rule.define_log_tag())

    def test_check_condition(self):
        file_path = self.__generate_tmp_file_path()
        mock_return = [1, 2, 3, 6, 7, 12]
        with patch('time.time', side_effect=mock_return):
            self.rule.generate_start_message()
            self.false(self.rule.check_condition("dummy", "dummy", "dummy"))
            self.__write_word(file_path, "hoge")
            self.false(self.rule.check_condition("dummy", "dummy", "dummy"))
            self.false(self.rule.check_condition("dummy", "dummy", "dummy"))
            self.__write_word(file_path, "stop")
            self.false(self.rule.check_condition("dummy", "dummy", "dummy"))
            self.true(self.rule.check_condition("dummy", "dummy", "dummy"))

    def test_generate_start_message(self):
        message = self.rule.generate_start_message()
        self.include(self.rule.TARGET_WARD, message)
        self.include(self.rule.monitor_file_path, message)
        self.include(str(self.rule.watch_interval), message)

    def test_generate_finish_message(self):
        file_path = self.__generate_tmp_file_path()
        msg = self.rule.generate_finish_message(5)
        self.include(str(5), msg)
        self.include(file_path, msg)

    def __generate_tmp_file_path(self):
        return os.path.join(os.path.dirname(__file__), "tmp_file_for_manual_interruption_test.tmp")

    def __write_word(self, filepath, word):
      with open(filepath, 'wb') as f: f.write(word)

class WatchIterationCountTest(BaseUnitTest):

    def setUp(self):
        self.rule = WatchIterationCount(target_count=100)

    def tearDown(self):
        release_capture()

    def test_define_log_tag(self):
        self.eq("Progress", self.rule.define_log_tag())

    def test_check_condition(self):
        self.false(self.rule.check_condition(99, "dummy", "dummy"))
        self.true(self.rule.check_condition(100, "dummy", "dummy"))
        self.true(self.rule.check_condition(101, "dummy", "dummy"))

    def test_generate_start_message(self):
        self.include(str(100), self.rule.generate_start_message())

    def test_generate_finish_message(self):
        with patch('time.time', side_effect=[1477394336.834469, 1477394338.815435]):
            self.rule.generate_start_message()
            message = self.rule.generate_finish_message(5)
            self.include(str(5), message)
            self.include(str(1), message)

    def test_message_for_lifecycle(self):
        rule = WatchIterationCount(target_count=2)
        mock_time_return = [
            1477462649.467603,
            1477462653.996855,
            1477462655.032283,
            1477462658.945142,
            1477462663.027276,
            1477462679.056145
        ]
        expected_calc_time = [1.0354280471801758, 4.082134008407593]
        expected_interrupt_return = [False, True]
        expected_total_time = 29.588541984558105

        with patch('time.time', side_effect=mock_time_return):
            rule.before_gpi_start("dummy", "dummy")
            for idx, expected in enumerate(expected_calc_time, start=1):
                capture = StringIO.StringIO()
                sys.stdout = capture
                rule.before_update("dummy", "dummy", "dummy")
                rule.after_update(idx, "dummy", "dummy")
                self.eq(expected_interrupt_return[idx-1], rule.interrupt_gpi(idx, "dummy", "dummy"))
                self.include("[Progress]", capture.getvalue())
                self.include("%.1f" % expected, capture.getvalue())
                self.include("%d /" % idx, capture.getvalue())
            rule.after_gpi_finish("dummy", "dummy")
            self.include("%d" % expected_total_time, capture.getvalue())

    def test_verbose_mode(self):
        capture = StringIO.StringIO()
        sys.stdout = capture
        rule = WatchIterationCount(target_count=100, verbose=0)
        rule.before_gpi_start("dummy", "dummy")
        rule.before_update(1, "dummy", "dummy")
        rule.after_update(1, "dummy", "dummy")
        rule.interrupt_gpi(1, "dummy", "dummy")
        rule.before_update(100, "dummy", "dummy")
        rule.after_update(100, "dummy", "dummy")
        rule.interrupt_gpi(100, "dummy", "dummy")
        rule.after_gpi_finish("dummy", "dummy")
        self.include("Completed", capture.getvalue())
        self.not_include("Finished", capture.getvalue())




from examples.ticktacktoe.task import TickTackToeTask
from tests.base_unittest import BaseUnitTest


class TickTackToeTaskTest(BaseUnitTest):
    """
    Board and bit-flg relation
    EX1.
        (O) first_player_board  = 110000000     OO-
        (X) second_player_board = 000000011  => ---
                                                -XX
    EX2.
        (O) first_player_board  = 101010101     OXO
        (X) second_player_board = 010101010  => XOX
                                                OXO
    """

    def setUp(self):
        self.task = TickTackToeTask()

    def test_generate_inital_state(self):
        state = self.task.generate_initial_state()
        self.eq((0,0), state)

    def test_terminal_state_check_when_initial_state(self):
        state = gen_initial_state()
        self.false(self.task.is_terminal_state(state))

    def test_terminal_state_check_when_draw(self):
        state = gen_draw_state()
        self.true(self.task.is_terminal_state(state))

    def test_terminal_state_check_when_vertical_win(self):
        state = gen_vertical_win_state()
        self.true(self.task.is_terminal_state(state))

    def test_terminal_state_check_when_horizontal_win(self):
        state = gen_horizontal_win_state()
        self.true(self.task.is_terminal_state(state))

    def test_terminal_state_check_when_diagonal_win(self):
        state = gen_diagonal_win_state()
        self.true(self.task.is_terminal_state(state))

    def test_transit_state_from_initial_state(self):
        state = gen_initial_state()
        state = self.task.transit_state(state, 1)
        state = self.task.transit_state(state, 2)
        state = self.task.transit_state(state, 256)
        expected = (257, 2)
        self.eq(expected, state)

    def test_generate_possible_actions_when_draw(self):
        state = gen_draw_state()
        expected = []
        self.eq(expected, self.task.generate_possible_actions(state))

    def test_generate_possible_actions_when_initial_state(self):
        state = gen_initial_state()
        expected = [1, 2, 4, 8, 16, 32, 64, 128, 256]
        self.eq(expected, self.task.generate_possible_actions(state))

    def test_generate_possible_actions_when_vertical_win(self):
        state = gen_vertical_win_state()
        expected = [1, 2, 8, 64]
        self.eq(expected, self.task.generate_possible_actions(state))

    def test_generate_possible_actions_when_horizontal_win(self):
        state = gen_horizontal_win_state()
        expected = [1, 2, 4, 8]
        self.eq(expected, self.task.generate_possible_actions(state))

    def test_generate_possible_actions_when_diagonal_win(self):
        state = gen_diagonal_win_state()
        expected = [2, 4, 8, 32]
        self.eq(expected, self.task.generate_possible_actions(state))

    def test_calculate_reward_when_initial_state(self):
        state = gen_initial_state()
        self.eq(TickTackToeTask.REWARD_NONE, self.task.calculate_reward(state))

    def test_calculate_reward_when_first_player_win(self):
        state = gen_diagonal_win_state()
        self.eq(TickTackToeTask.REWARD_WIN, self.task.calculate_reward(state))

    def test_calculate_reward_when_first_player_lose(self):
        state = gen_diagonal_win_state()[::-1]
        self.eq(TickTackToeTask.REWARD_LOSE, self.task.calculate_reward(state))

    def test_is_first_player_flg(self):
        task = TickTackToeTask(is_first_player=False)
        state = gen_diagonal_win_state()
        self.eq(TickTackToeTask.REWARD_LOSE, task.calculate_reward(state))


def gen_initial_state():
    """
    ---
    ---
    ---
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("000000000")
    second_player_board = bin2i("000000000")
    state = (first_player_board, second_player_board)
    return state

def gen_draw_state():
    """
    OXO
    OXO
    XOX
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("010101101")
    second_player_board = bin2i("101010010")
    state = (first_player_board, second_player_board)
    return state

def gen_vertical_win_state():
    """
    OX-
    OX-
    O--
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("100100100")
    second_player_board = bin2i("010010000")
    state = (first_player_board, second_player_board)
    return state

def gen_horizontal_win_state():
    """
    OOO
    XX-
    ---
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("111000000")
    second_player_board = bin2i("000110000")
    state = (first_player_board, second_player_board)
    return state

def gen_diagonal_win_state():
    """
    OXX
    -O-
    --O
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("100010001")
    second_player_board = bin2i("011000000")
    state = (first_player_board, second_player_board)
    return state




import examples.ticktacktoe.helper as H
from examples.ticktacktoe.task import TickTackToeTask
from tests.base_unittest import BaseUnitTest


class TickTackToeHelperTest(BaseUnitTest):

    def test_visualize_board_when_empty(self):
        bin2i = lambda b: int(b, 2)
        first_player_board = bin2i("000000000")
        second_player_board = bin2i("000000000")
        state = (first_player_board, second_player_board)
        expected = "- - -\n- - -\n- - -"
        self.eq(expected, H.visualize_board(state))

    def test_visualize_board_when_not_empty(self):
        bin2i = lambda b: int(b, 2)
        first_player_board = bin2i("110000000")
        second_player_board = bin2i("000000011")
        state = (first_player_board, second_player_board)
        expected = "O O -\n- - -\n- X X"
        self.eq(expected, H.visualize_board(state))

    def test_feature_construction(self):
        bin2i = lambda b: int(b, 2)
        first_player_board = bin2i("000000100")
        second_player_board = bin2i("000000010")
        state = (first_player_board, second_player_board)
        action = 1
        features = H.construct_features(TickTackToeTask(), state, action)
        self.eq(18, len(features))
        self.eq([1,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0], features)




import random

from kyoka.utils import build_not_implemented_msg


def choose_best_action(task, value_function, state, rand=None):
    """Calculate greedy action by using passed value function

    Args:
        task : used to generate possible actions in passed state
        value_function : used to calculate value of each action
        state : greedy action is selected from possible actions of this state
    """
    rand = rand if rand else random
    actions = task.generate_possible_actions(state)
    Q_value_for_actions = [value_function.predict_value(state, action) for action in actions]
    max_Q_value = max(Q_value_for_actions)
    Q_act_pair = zip(Q_value_for_actions, actions)
    best_actions = [act for Q_value, act in Q_act_pair if max_Q_value == Q_value]
    best_action = rand.choice(best_actions)
    return best_action

class BasePolicy(object):
    """Base class for creating new Policy as you like.

    Policy is used to choose action from passed state by following some logic.
    """

    def choose_action(self, task, value_function, state):
        """
        Returns:
            action: an action choosed from task.generate_possible_actions(state) by some logic
        """
        err_msg = build_not_implemented_msg(self, "choose_action")
        raise NotImplementedError(err_msg)

class GreedyPolicy(BasePolicy):
    """Choose action which has maximum value in the state"""

    def __init__(self, rand=None):
        """Args:
            rand : rand.choice is used when multiple best action found.
                   Default is random module of python module.
        """
        self.rand = rand if rand else random

    def choose_action(self, task, value_function, state):
        return choose_best_action(task, value_function, state, self.rand)

class EpsilonGreedyPolicy(BasePolicy):
    """Choose explore action in probability epsilon else choose best action

    If you want to decaly epsilon (most of the case it's good idea) during training,
    call "set_eps_annealing" before start training.

    Properties:
        eps : the probability to select explore action
        rand : used when multiple actions are candidate of choice
    """

    def __init__(self, eps=0.05, rand=None):
        self.eps = eps
        self.rand = rand if rand else random
        self.do_annealing = False

    def choose_action(self, task, value_function, state):
        actions = task.generate_possible_actions(state)
        best_action = choose_best_action(task, value_function, state, self.rand)
        probs = self.__calc_select_probability(best_action, actions)
        selected_action_idx = self.__roulette(probs)
        return actions[selected_action_idx]

    def set_eps_annealing(self, initial_eps, final_eps, anneal_duration):
        """Call this method enables epsilon annealing in training

        Args:
            initial_eps: Override initial value of epsilon passed in constructor
            final_eps: Epsilon is decayed from initial_eps to final_eps through anneal_duration
            anneal_duration: Epsilon is uniformly decayed through this duration
        """
        self.do_annealing = True
        self.eps = initial_eps
        self.min_eps = final_eps
        self.anneal_step = (initial_eps - final_eps) / anneal_duration

    def anneal_eps(self):
        self.eps = max(self.min_eps, self.eps - self.anneal_step)

    def __calc_select_probability(self, best_action, actions):
        e = self.eps / len(actions)
        bonus = 1 - self.eps
        calc_prob = lambda action: e + bonus if action == best_action else e
        return [calc_prob(action) for action in actions]

    def __roulette(self, probs):
        hit_idx = -1
        prob_sum = 0
        dart = self.rand.random()
        for idx, prob in enumerate(probs):
            prob_sum += prob
            if dart < prob_sum:
                hit_idx = idx
                break
        return hit_idx




import os

from nose.tools import raises

from kyoka.value_function import BaseTabularActionValueFunction, BaseApproxActionValueFunction
from tests.base_unittest import BaseUnitTest
from tests.utils import generate_tmp_dir_path, setup_tmp_dir, teardown_tmp_dir


class BaseTabularActionValueFunctionTest(BaseUnitTest):

    def setUp(self):
        self.func = self.TestImpl()
        self.func.setup()

    def tearDown(self):
        tmp_file_name = "hoge_table_action_value_function_data.pickle"
        teardown_tmp_dir(__file__, [tmp_file_name])

    @raises(NotImplementedError)
    def test_backup(self):
        BaseTabularActionValueFunction().backup("dummy", "dummy", "dummy", "dummy")

    @raises(NotImplementedError)
    def test_generate_initial_table(self):
        BaseTabularActionValueFunction().generate_initial_table()

    @raises(NotImplementedError)
    def test_fetch_value_from_table(self):
        BaseTabularActionValueFunction().fetch_value_from_table("dummy", "dummy", "dummy")

    @raises(NotImplementedError)
    def test_insert_value_into_table(self):
        BaseTabularActionValueFunction().insert_value_into_table("dummy", "dummy", "dummy", "dummy")

    def test_predict_value(self):
        state, action = 0, 1
        self.func.table[state][action]= 1
        self.eq(1, self.func.predict_value(state, action))

    def test_insert_value_into_table(self):
        state, action = 0, 1
        self.eq(0, self.func.predict_value(state, action))
        self.func.insert_value_into_table(self.func.table, state, action, 1)
        self.eq(1, self.func.predict_value(state, action))

    def test_store_and_restore_table(self):
        setup_tmp_dir(__file__)
        dir_path = generate_tmp_dir_path(__file__)
        file_path = os.path.join(dir_path, "hoge_table_action_value_function_data.pickle")
        state, action = 0, 1
        self.func.insert_value_into_table(self.func.table, state, action, 1)
        self.eq(1, self.func.predict_value(state, action))
        self.false(os.path.exists(file_path))
        self.func.save(dir_path)
        self.true(os.path.exists(file_path))
        self.func = self.TestImpl()
        self.func.load(dir_path)
        self.eq(1, self.func.predict_value(state, action))

    class TestImpl(BaseTabularActionValueFunction):

        def generate_initial_table(self):
            return [[0 for j in range(2)] for i in range(1)]

        def fetch_value_from_table(self, table, state, action):
            return table[state][action]

        def insert_value_into_table(self, table, state, action, new_value):
            table[state][action] = new_value

        def define_save_file_prefix(self):
            return "hoge"


class BaseApproxActionValueFunctionTest(BaseUnitTest):

    def setUp(self):
        self.func = BaseApproxActionValueFunction()

    def test_predoct_value(self):
        func = self.TestImpl()
        self.eq("predict:hogefuga", func.predict_value("hoge", "fuga"))

    def test_backup(self):
        func = self.TestImpl()
        func.backup("hoge", "fuga", "dummy", "dummy")
        self.eq("backup:hogefuga", func.memo)

    @raises(NotImplementedError)
    def test_error_msg_when_not_implement_construct_features(self):
        self.func.construct_features("dummy", "dummy")

    @raises(NotImplementedError)
    def test_error_msg_when_not_implement_approx_predict_value(self):
        self.func.approx_predict_value("dummy")

    @raises(NotImplementedError)
    def test_error_msg_when_not_implement_approx_backup(self):
        self.func.approx_backup("dummy", "dummy", "dummy")

    class TestImpl(BaseApproxActionValueFunction):

        def construct_features(self, state, action):
            return state + action

        def approx_predict_value(self, features):
            return "predict:%s" % features

        def approx_backup(self, features, backup_target, alpha):
            self.memo = "backup:%s" % features

